package com.svv.cbr;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import com.svv.cbr.cases.*;

/**
 * Uses methods from the TrafficSituation class to retrieve cases, and process the retrived cases
 * @author Fredrick, yvind
 *
 */
public class Cbr{
	private TrafficSituation trafficSituation;
	
	public Cbr()
	{
		try
		{
			trafficSituation = new TrafficSituation();
		}catch(Exception e){
			System.out.println("CBR ERROR1: " + e.getMessage());
		}
	}
	
	/**
	 * Initializes the retrieval of cases that matches the query
	 * @param queryCase The query case built by the system
	 * @return A set with all retrieved cases, if successful, or null otherwise
	 */
	public ArrayList<TrafficSituationCase> getMatchingCases(TrafficSituationCase queryCase)
	{
		try
		{
			return trafficSituation.getCasesMatchingQuery(queryCase);
			
		}catch(Exception e){
			System.out.println("CBR ERROR2: " + e.getMessage());
		}
		return null;
	}
	/**
	 * Runs the method for retrieving the cases that matches the query, and runs the method for finding the best one in the set of retrieved cases.
	 * @param queryCase The query case that built by the system
	 * @return The best solution to the problem, if successful, or null otherwise
	 */
	public String getSolutionCase(TrafficSituationCase queryCase)
	{
		try
		{
			ArrayList<TrafficSituationCase> trafficSituationCases = trafficSituation.getCasesMatchingQuery(queryCase);
			return findBestTrafficSituationCase(trafficSituationCases);
		} catch(Exception e) {
			System.out.println("CBR ERROR2: " + e.getMessage());
		}
		return null;
	}
	/**
	 * Finds the best case out of a list of cases by looking at their similarity to the query case
	 * @param trafficSituationCases An ArrayList containing retrieved cases
	 * @return The best solution to the problem
	 */
	public String findBestTrafficSituationCase(ArrayList<TrafficSituationCase> trafficSituationCases)
	{
		HashMap<String, Integer> solutionList = new HashMap<>();
		for(TrafficSituationCase trafficCase : trafficSituationCases)
		{
			String solution = trafficCase.getSolution();
			if(!solutionList.containsKey(solution))
			{
				solutionList.put(solution, 1);
			}
			else
			{
				int value = (int)solutionList.get(solution) + 1;
				solutionList.put(solution, value);
			}
		}
		return getMaxValue(solutionList);
	}
	/**
	 * Chooses the best solution in a list of retrieved cases based on a majority vote.
	 * @param solutionList All retrieved cases
	 * @return A case with the most common solution among all retrieved cases
	 */
	private String getMaxValue(HashMap<String, Integer> solutionList)
	{
		int maxValue = 0;
		String solution = "";
		for(Map.Entry<String, Integer> entry: solutionList.entrySet())
		{
			if(maxValue < entry.getValue())
			{
				maxValue = entry.getValue();
				solution = entry.getKey();
			}
		}
		return solution;
	}
	
	/**
	 * Is activated if an expert decides to save a query case as a new case in the case base. Stores the case in the case base.
	 * @param newTrafficSituationCase The query case
	 */
	public void retain(TrafficSituationCase newTrafficSituationCase)
	{
		try
		{
			trafficSituation.saveCaseToXML(newTrafficSituationCase);
		}catch(Exception e)
		{
			System.out.println("Cbr error saveCaseToXML: " + e.getMessage());
		}
		
	}
}
