package com.svv.cbr;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import com.svv.cbr.cases.TrafficSituationCase;
import com.svv.util.AddToXml;
import com.svv.util.Config;

import de.dfki.mycbr.core.Project;
import de.dfki.mycbr.core.casebase.Case;
import de.dfki.mycbr.core.casebase.Instance;
import de.dfki.mycbr.core.model.Concept;
import de.dfki.mycbr.core.model.FloatDesc;
import de.dfki.mycbr.core.model.SymbolDesc;
import de.dfki.mycbr.core.retrieval.Retrieval;
import de.dfki.mycbr.core.similarity.Similarity;

/**
 * Uses myCBR development tool to create the case base and retrieve cases
 * @author Fredrick, yvind
 *
 */
public class TrafficSituation{
	private SymbolDesc proposedTransitionTime;
	private Concept trafficSituation;
	private Project projectTrafficSituation;
	
	public TrafficSituation() throws Exception
	{
		projectTrafficSituation = new Project(Config.TRAFFIC_SITUATION);
		
		HashMap<String, Concept> hm = projectTrafficSituation.getAllSubConcepts();
		trafficSituation = hm.get("TrafficSituation");
		
		proposedTransitionTime = (SymbolDesc)trafficSituation.getAttributeDesc("ProposedTransitionTime");
	}
	
	/**
	 * Retrieves the cases that matches the query case.
	 * @param queryCase The query case created by the system based on available traffic data
	 * @return A list of retrieved cases
	 * @throws Exception
	 */
	public ArrayList<TrafficSituationCase> getCasesMatchingQuery(TrafficSituationCase queryCase) throws Exception
	{
		Retrieval r = doQuery(queryCase);
		
		ArrayList<TrafficSituationCase> solutionTrafficCases = new ArrayList<>();
		double maxSimilarity = 0;
		for (Map.Entry<Case, Similarity> entry: r.entrySet()) {
			double currentSimilarity = entry.getValue().getValue();
			if(currentSimilarity > maxSimilarity)
			{
				maxSimilarity = currentSimilarity;
				solutionTrafficCases.clear();
				solutionTrafficCases.add(createTrafficSituationCase(entry.getKey(), currentSimilarity));
			}
			else if(currentSimilarity == maxSimilarity)
			{
				solutionTrafficCases.add(createTrafficSituationCase(entry.getKey(), currentSimilarity));
			}
		}
		return solutionTrafficCases;
	}
	
	/**
	 * Converts a case from the type Case (myCBR) to the type TrafficSituationCase. It also sets the similarity of the case.
	 * @param c The case as a Case type
	 * @param similarity 
	 * @return A converted case of type TrafficSituationCase
	 */
	private TrafficSituationCase createTrafficSituationCase(Case c, double similarity)
	{
		String trafficFlow = c.getAttForDesc(trafficSituation.getAttributeDesc("TrafficFlow")).getValueAsString();
		boolean vehicleDetected = Boolean.parseBoolean(c.getAttForDesc(trafficSituation.getAttributeDesc("VehicleDetected")).getValueAsString());
		String speedOfDetectedVehicle = c.getAttForDesc(trafficSituation.getAttributeDesc("SpeedOfDetectedVehicle")).getValueAsString();
		String positiveIntentions = c.getAttForDesc(trafficSituation.getAttributeDesc("PositiveIntentions")).getValueAsString();
		String timeWaited = c.getAttForDesc(trafficSituation.getAttributeDesc("TimeWaited")).getValueAsString();
		String pedestrianType = c.getAttForDesc(trafficSituation.getAttributeDesc("PedestrianType")).getValueAsString();
		String solutionProposedTransitionTime = c.getAttForDesc(proposedTransitionTime).getValueAsString();
		
		TrafficSituationCase newCase = new TrafficSituationCase(trafficFlow, vehicleDetected, speedOfDetectedVehicle, positiveIntentions, timeWaited, pedestrianType);
		newCase.setSimilarity(similarity);
		newCase.setSolutionProposedTransitionTime(solutionProposedTransitionTime);
		return newCase;
	}
	/**
	 * Runs the retrieval (myCBR)
	 * @param queryCase The query case built by the system
	 * @return The results of the retrieval
	 * @throws Exception
	 */
	private Retrieval doQuery(TrafficSituationCase queryCase) throws Exception
	{
		Retrieval r = new Retrieval(trafficSituation);
		Instance q = r.getQuery();
		q.addAttribute("TrafficFlow", queryCase.getTrafficFlow());
		q.addAttribute("VehicleDetected", queryCase.getVehicleDetected());
		q.addAttribute("SpeedOfDetectedVehicle", queryCase.getSpeedOfDetectedVehicle());
		q.addAttribute("PositiveIntentions", queryCase.getPositiveIntentions());
		q.addAttribute("TimeWaited", queryCase.getTimeWaited());
		q.addAttribute("PedestrianType", queryCase.getPedestrianType());
		r.start();
		return r;
	}
	/**
	 * Saves a case to XML (case base is stored in XML format).
	 * This is used for when experts want to update the case base, by adding the query case to the case base.
	 * @param trafficSituationCase The query case that the user wants to add
	 * @throws Exception
	 */
	public void saveCaseToXML(TrafficSituationCase trafficSituationCase) throws Exception
	{
		AddToXml addToXml = new AddToXml(trafficSituationCase);
		addToXml.saveInstanceToXML();
		projectTrafficSituation = new Project(Config.TRAFFIC_SITUATION);
		HashMap<String, Concept> hm = projectTrafficSituation.getAllSubConcepts();
		trafficSituation = hm.get("TrafficSituation");
		proposedTransitionTime = (SymbolDesc)trafficSituation.getAttributeDesc("ProposedTransitionTime");
	}
}