package com.svv.util;

import java.util.ArrayList;

import com.svv.cbr.cases.TrafficSituationCase;

/**
 * Methods to handle different types of query requests.
 * Also used to handle request from either the real system or the test system.
 * @author Fredrick, yvind
 *
 */
public class QueryBuilder {
	
	public static int TEST_QUERIES = 1;
	public static int NORMAL = 0;
	
	private long pausedTime = -1;
	private TrafficEventHandler trafficEventHandler;
	Query query;
	
	public QueryBuilder(TrafficEventHandler trafficEventHandler)
	{
		query = trafficEventHandler;
		this.trafficEventHandler = trafficEventHandler;
	}
	
	public TrafficSituationCase getTrafficQuery()
	{
		return query.getTrafficQuery();
	}
	/**
	 * Gets the time pedestrians have waited for this query.
	 * @return The time waited
	 */
	public double getTimeWaited()
	{
		return query.getTimeWaited();
	}
	
	public void clear()
	{
		query.clear();
	}
	/**
	 * Changes between normal mode (where the system gets its cases from information from a simulator or similar) and test mode (where the query cases are given to the system)
	 * @param mode Either 0 or 1, 0 = normal mode, 1 = test mode
	 */
	public void setMode(int mode)
	{
		if(NORMAL == mode)
		{
			query = trafficEventHandler;
		}
		else if(TEST_QUERIES == mode)
		{
			query = new TestQueries();
		}
	}
	
	public void pause()
	{
		if(pausedTime == -1)
		{
			pausedTime = System.currentTimeMillis();
		}
	}
	
	public void start()
	{
		long totalPauseTime = System.currentTimeMillis() - pausedTime;
		query.addPausedTime(totalPauseTime);
		pausedTime = -1;
	}
	
	/**
	 * Used when the TEST_QUERIES mode is active.
	 * Read query cases from a file.
	 * @author Fredrick
	 *
	 */
	private class TestQueries implements Query
	{
		private int currentCase = 0;
		private long startTime = -1;
		
		private ArrayList<String[]> trafficCases;
		private CaseBuilder caseBuilder;
		
		public TestQueries()
		{
			caseBuilder = new CaseBuilder();
			trafficCases = ReadTrafficData.getTestCaseBase();
		}
		
		@Override
		public TrafficSituationCase getTrafficQuery() {
			if(startTime == -1)
			{
				startTime = System.currentTimeMillis();
			}
			if(currentCase > trafficCases.size())
			{
				clear();
				return null;
			}
			String[] trafficCase  = trafficCases.get(currentCase);
			
			return caseBuilder.createCase(Double.parseDouble(trafficCase[0]), Boolean.parseBoolean(trafficCase[1]), Double.parseDouble(trafficCase[2]), 
					Double.parseDouble(trafficCase[3]), getTimeWaited(), trafficCase[4]);
		}

		@Override
		public double getTimeWaited() {
			return (System.currentTimeMillis() - startTime)/1000.0;
		}

		@Override
		public void clear() {
			if(currentCase < trafficCases.size()-1)
			{
				startTime  = -1;
				currentCase++;
			}
			else
			{
				setMode(QueryBuilder.NORMAL);
			}
		}

		@Override
		public void addPausedTime(long timeMillis) {
			if(startTime != -1)
			{
				startTime += timeMillis;
			}
		}
		
	}
}
