package com.svv.util;

import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;

import org.osgi.service.event.Event;
import org.osgi.service.event.EventHandler;

import com.svv.cbr.cases.TrafficSituationCase;
import com.svv.trafficcbr.Activator;

/**
 * Receive data from different sensors that is stored in this class.
 * It also contains methods to get a query case, built from the stored data.
 * @author Fredrick, yvind
 *
 */
public class TrafficEventHandler implements EventHandler, KeyListener, Query{
	private final double TRIGGER_DISTANCE = 70;
	
	private double trafficFlow = 10.0, speedOfDetectedVehicle = -1, positiveIntentions, keyEventPositiveIntentions;
	private long startTime = -1, timeVehicleDetected = -1;
	private String pedestrianType = "Normal";
	private CaseBuilder caseBuilder;
	private Activator activator;
	
	private final double MINUTES = 5;
	private long tmpTimeStamp = -1;
	private int counter = (int)trafficFlow;
	
	public TrafficEventHandler(CaseBuilder caseBuilder, Activator activator)
	{
		this.caseBuilder = caseBuilder;
		this.activator = activator;
	}
	
	/**
	 * Use all data stored from the sensors and build a query case
	 */
	@Override
	public TrafficSituationCase getTrafficQuery() 
	{
		double p = getIntention();
		if(p > 0.0) {
			double currentTime = (System.currentTimeMillis() - startTime)/1000.0;
			if(checkVehicleDetected())
			{
				return caseBuilder.createCase(trafficFlow, true, speedOfDetectedVehicle, p, currentTime, pedestrianType); 
			}
			else
			{
				TrafficSituationCase trafficCase = caseBuilder.createCase(trafficFlow, false, -1, p, currentTime, pedestrianType); 
				speedOfDetectedVehicle = -1;
				return trafficCase;
			}
			
		}
		return null;
	}
	
	/**
	 * Use the speed of the vehicle, the time the vehicle was detected, the distance between the loop and traffic light (TRIGGER_DISTANCE)
	 * and the current time to calculate if a vehicle is detected.
	 * @return If a vehicle is detected true is returned
	 */
	private boolean checkVehicleDetected()
	{
		if(timeVehicleDetected == -1)
		{
			return false;
		}
		double timePassed = (System.currentTimeMillis() - timeVehicleDetected)/1000.00;
		double vehicleDetectedTime = 0;
		if(speedOfDetectedVehicle > 0)
		{
			vehicleDetectedTime = TRIGGER_DISTANCE/(speedOfDetectedVehicle/3.6);
			
		}
		else
		{
			return false;
		}
		
		if(timePassed > vehicleDetectedTime)
		{
			timeVehicleDetected = -1;
			return false;
		}
		else
		{
			return true;
		}
	}
	
	/**
	 * Reset all values back to standard
	 */
	public synchronized void clear()
	{
		startTime = -1;
		positiveIntentions = 0;
		keyEventPositiveIntentions = 0;
		pedestrianType = "Normal";
	}

	/**
	 * Called every time a vehicle pass the loops in either directions. 
	 * Calculate vehicles that passed the last five minutes.
	 */
	private void calculateTrafficFlow()
	{
		double timePassed = ((System.currentTimeMillis() - tmpTimeStamp)/1000)/60;
		if(timePassed < MINUTES)
		{
			counter++;
		}
		else
		{
			trafficFlow = counter;
			counter = 1;
			tmpTimeStamp = System.currentTimeMillis();
		}
	}
	
	/**
	 * Events from other systems; the intention system (org/npra/ven/KINECT/) and SCANeR Studio (org/npra/ven/SCANeR)
	 */
	@Override
	public void handleEvent(Event eve)
	{
		try{
			String[] propertyNames=eve.getPropertyNames();
			for(int i =0;i<propertyNames.length;i++)
			{
				if(propertyNames[i].contains("org/npra/ven/KINECT/INTENTION"))
				{
					Object propertyValue = (Object) eve.getProperty(propertyNames[i]);
					setIntention((double)propertyValue);
					System.out.println("TrafficEventHandler intentions: "+getIntention());
					if(getIntention() == 0.0)
					{
						startTime = -1;
					} else if(startTime == -1) {
						startTime = System.currentTimeMillis();
					}
				}
				else if(propertyNames[i].contains("org/npra/ven/KINECT/TYPE"))
				{
					String pedestrianType = (String)eve.getProperty(propertyNames[i]);
					if(this.pedestrianType.equals("Slow") && pedestrianType.equals("SG"))
					{
						this.pedestrianType = "SlowGroup";
					}
					else if(this.pedestrianType.equals("Normal"))
					{
						if(pedestrianType.equals("S"))
						{
							this.pedestrianType = "Slow";
						}
						else if(pedestrianType.equals("F"))
						{
							this.pedestrianType = "Fast";
						}
						else if(pedestrianType.equals("SG"))
						{
							this.pedestrianType = "SlowGroup";
						}
					}
					else if(this.pedestrianType.equals("Fast"))
					{
						if(pedestrianType.equals("N"))
						{
							this.pedestrianType = "Normal";
						}
						else if(pedestrianType.equals("S"))
						{
							this.pedestrianType = "Slow";
						}
						else if(pedestrianType.equals("SG"))
						{
							this.pedestrianType = "SlowGroup";
						}
					}
				}
				else if(propertyNames[i].contains("org/npra/ven/SCANeR"))
				{
					Object propertyValue = (Object) eve.getProperty(propertyNames[i]);
					switch(propertyNames[i].charAt(propertyNames[i].length()-1))
					{
						case('1'):
							System.out.println("Traffic flow: "+trafficFlow);
							double tmpSpeed = (double)propertyValue;
							calculateTrafficFlow();
							timeVehicleDetected = System.currentTimeMillis();
							if(tmpSpeed > 0)
							{
								speedOfDetectedVehicle = tmpSpeed;
							}
							else
							{
								speedOfDetectedVehicle = 50.0;
							}

							break;
						case('0'):
							break;
					}
				}
			}	

		}catch(Exception e){
			System.out.println("Error:" + e.getMessage());
		}
	}
	
	public synchronized void setIntention(double intention)
	{
		positiveIntentions = intention;
	}
	
	public synchronized double getIntention() 
	{
		return positiveIntentions + keyEventPositiveIntentions;
	}
	
	@Override
	public double getTimeWaited()
	{
		return (System.currentTimeMillis() - startTime)/1000.0;
	}
	
	@Override
	public void addPausedTime(long timeMillis) {
		if(startTime != -1)
		{
			startTime += timeMillis;
		}
	}
	
	@Override
	public void keyTyped(KeyEvent e) {}
	
	/**
	 * Key event listener that is binded to the GUI.
	 * Makes it possible to send different types of pedestrians and pause the system.
	 */
	@Override
	public void keyPressed(KeyEvent e) {
		//s = slow pedestrian
		if(e.getKeyChar() == 's')
		{
			pedestrianType = "Slow";
			if(startTime == -1)
			{
				keyEventPositiveIntentions = 1.0;
				startTime = System.currentTimeMillis();
			}
			else
			{
				keyEventPositiveIntentions++;
			}
		}
		//f = fast pedestrian
		else if (e.getKeyChar() == 'f') 
		{
			pedestrianType = "Fast";
			if(startTime == -1)
			{
				keyEventPositiveIntentions = 1.0;
				startTime = System.currentTimeMillis();
			}
			else
			{
				keyEventPositiveIntentions++;
			}
		}
		else if(e.getKeyChar() == 'n')
		{
			pedestrianType = "Normal";
			if(startTime == -1)
			{
				keyEventPositiveIntentions = 1.0;
				startTime = System.currentTimeMillis();
			}
			else
			{
				keyEventPositiveIntentions++;
			}
		}
		else if(e.getKeyChar() == 'c')
		{
			clear();
		}
		else if(e.getKeyChar() == 'e')
		{
		}
		else if(e.getKeyChar() == 'p')
		{
			activator.pauseThread();
		}
	}
	@Override
	public void keyReleased(KeyEvent e) {}
}