/*
 * myCBR License 2.0
 *
 * Copyright (c) 2010
 * Thomas Roth-Berghofer, Armin Stahl & German Research Center for Artificial
 * Intelligence DFKI GmbH
 * All rights reserved.
 *
 * myCBR is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Since myCBR uses some modules, you should be aware of their licenses for
 * which you should have received a copy along with this program, too.
 *
 * endOfLic */
package de.dfki.mycbr.core;

import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Observable;

import de.dfki.mycbr.core.casebase.Case;
import de.dfki.mycbr.core.casebase.Instance;
import de.dfki.mycbr.core.model.AttributeDesc;
import de.dfki.mycbr.core.model.Concept;

/**
 * Class DefaultCasebase. Represents cases as a c attribute. Therefore, uses
 * myCBR's default classes for internal case representation contained in the
 * package <code>de.dfki.mycbr.core.casebase</code>.
 *
 * @author myCBR Team
 * @since myCBR 3.0.0
 */
public final class DefaultCaseBase extends Observable implements ICaseBase {

    /**
     * Maps c description to a list of c attributes. For a faster access of
     * cases belonging to a given c
     */
    private HashMap<Concept, List<Case>> cases;

    /**
     */
    private String name;

    /**
     */
    private Project prj;

    /**
     * Initializes this with the given list of concept descriptions. The list
     * specifies the highest concepts in the inheritance hierarchy. To create a
     * DefaultCase base call {@link Project#createDefaultCB(String)}.
     *
     * @param cbName
     *            the name of the case base
     * @param p
     *            the project this cb should belong to
     * @throws Exception
     *             if name is null or empty or if the given project is null or
     *             already has a case base with this name
     */
    DefaultCaseBase(final Project p, final String cbName) throws Exception {
        cases = new HashMap<Concept, List<Case>>();
        if (cbName == null || cbName.trim().equals("")) {
            throw new Exception("Cannot add case base with empty name");
        } else if (p.getCB(cbName) != null) {
            throw new Exception("Project already has a case base with name \""
                    + cbName + "\"");
        }
        this.name = cbName;
        this.prj = p;
    }

    /**
	 * @param cbName
     *            the name of the case base
     * @param p
     *            the project this cb should belong to
     * @param count
     * 				the number of cases expected for this case base
     * 
     * @throws Exception
     *             if name is null or empty or if the given project is null or
     *             already has a case base with this name
	 */
	public DefaultCaseBase(Project p, String cbName, int count) throws Exception {
		cases = new HashMap<Concept, List<Case>>(count);
        if (cbName == null || cbName.trim().equals("")) {
            throw new Exception("Cannot add case base with empty name");
        } else if (p.getCB(cbName) != null) {
            throw new Exception("Project already has a case base with name \""
                    + cbName + "\"");
        }
        this.name = cbName;
        this.prj = p;
	}

	/**
     * Gets the current cases known for the given description. Be aware that you
     * do not get the cases that are known for sub concepts of the given
     * description, call {@link #getAllCases(Concept)} instead.
     *
     * @param c the concept whose known cases should be returned
     * @return cases for the given description
     */
    public List<Case> getCases(final Concept c) {
        List<Case> list = cases.get(c);
        if (list == null) {
            list = new LinkedList<Case>();
        }
        return list;
    }

    /**
     * Gets all cases known for the given description. These include the cases
     * known for sub descriptions of the given c. If you only want to have the
     * proper cases call {@link #getCases(Concept)}.
     *
     * @param c the concept for which all cases should be returned
     * @return all cases for the given description
     */
    public List<Case> getAllCases(final Concept c) {
        List<Case> allCasesForConcept = cases.get(c);
        if (allCasesForConcept == null) {
            allCasesForConcept = new LinkedList<Case>();
        }
        for (Concept sub : c.getSubConcepts().values()) {
            allCasesForConcept.addAll(getAllCases(sub));
        }
        return allCasesForConcept;
    }

    /**
     * Adds a new case to the list of known cases for the given description. If
     * there is no case known for this description yet, a new list is created
     * and the new case is added to this list. Returns the newly created case.
     *
     * @param i the new case's instance.
     * @param id the ID of the new case
     * @return the newly created case
     * @throws Exception if there is a case with this id
     */
    public Case addCase(final Instance i, final String id)
                                                            throws Exception {
        if (!containsCase(id)) {
            Case c = new Case(this, i, id);
            if (!cases.containsKey(i.getConcept())) {
                cases.put(i.getConcept(), new LinkedList<Case>());
            }
            cases.get(i.getConcept()).add(c);
            c.addObserver(this);
            return c;
        }

        throw new Exception("Case base already contains a case with id \""
                + id + "\"");
    }

    /**
     * Removes the given c attribute from the list of known cases. Returns true,
     * if att has really been a case and if this case has been removed
     * successfully, false otherwise.
     *
     * @param c the case to be removed
     * @return true, if the case has be removed successfully, false otherwise
     */
    public boolean removeCase(final Case c) {
        Concept concept = c.getInstance().getConcept();
        List<Case> atts = cases.get(concept);
        boolean contained = false;
        if (atts != null) {
            contained = atts.remove(c);
        }
        if (contained) {
            setCaseUnknown(c);
        }
        return contained;
    }

    /**
     * Returns the unique name of this case base.
     *
     * @return the unique name of this case base
     */
    @Override
    public String getName() {
        return name;
    }

    /**
     * Sets the name of this case base to the given name.
     * This method should not be called when name
     * is not unique.
     *
     * @param cbName the new name of this case base
     * @throws Exception if name is null or empty or there is a case base with
     *  this name in the corresponding project
     */
    @Override
    public void setName(final String cbName) throws Exception {
        if (cbName == null || cbName.trim().equals("")) {
            throw new Exception("Cannot give a case base an empty name!");
        }
        ICaseBase cbTMP = prj.getCaseBases().get(cbName);
        if (cbTMP != null) {
            // remove this casebase from the list
            prj.deleteCaseBase(this.name);
            this.name = cbName;
            prj.addCaseBase(this);
            setChanged();
            notifyObservers();
        } else {
            throw new Exception("Case base with name \"" + name + "\" "
                    + "already exists in project \"" + prj.getName() + "\"");
        }
    }

    /**
     * Tells whether this case base contains a case with the given name.
     *
     * @param caseName the name of the case to be checked for
     * @return true, if there is a case with the given name, false otherwise.
     */
    @Override
    public boolean containsCase(final String caseName) {
        if (caseName == null) {
            return false;
        }
        for (List<Case> casez : cases.values()) {
            for (Case c : casez) {
                if (caseName.equals(c.getInstance().getName())) {
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * Removes the first case with the given name. If there is no case with the
     * given name, does nothing.
     *
     * @param caseName the name of the case to be removed
     */
    @Override
    public void removeCase(final String caseName) {
        for (List<Case> casez : cases.values()) {
            for (Case c : casez) {
                if (caseName.equals(c.getInstance().getName())) {
                    casez.remove(c);
                    setCaseUnknown(c);
                    return;
                }
            }
        }
    }

    /**
     * Removes the first case with the given name. If there is no case with the
     * given name, does nothing.
     *
     * @param concept the concept the given case is associated to
     * @param c the case to be removed from this
     */
    public void removeCase(final Concept concept, final Case c) {
        List<Case> casez = cases.get(concept);
        if (casez != null) {
            boolean contained = casez.remove(c);
            if (contained) {
                setCaseUnknown(c);
            }
        }
    }

    /**
     * Cleans the given case which is no longer used.
     *
     * @param c the case which has been removed
     */
    private void setCaseUnknown(final Case c) {
        c.deleteObservers();
        c.getInstance().removeCase(c);
        setChanged();
        notifyObservers();
    }

    /**
     * Returns a list of all cases in this case base.
     *
     * @return all cases for this case base
     */
    @Override
    public Collection<Case> getCases() {
        Collection<Case> result = new LinkedList<Case>();
        for (List<Case> l : cases.values()) {
            result.addAll(l);
        }
        return result;
    }

    /*
     * (non-Javadoc)
     *
     * @see
     * de.dfki.mycbr.core.ICaseBase#removeCasesOfConcept(de.dfki.mycbr.core.
     * model.Concept)
     */
    @Override
    /**
     * Removes all cases of the given concept
     *
     * @param c the concept whose cases should be removed.
     */
    public void removeCasesOfConcept(final Concept c) {
        List<Case> cs = cases.get(c);
        if (cs != null) {
            for (Case ca : cs) {
                setCaseUnknown(ca);
            }
        }
        cases.remove(c);
    }

    /*
     * (non-Javadoc)
     *
     * @see de.dfki.mycbr.core.ICaseBase#clean(de.dfki.mycbr.core.model.Concept,
     * de.dfki.mycbr.core.model.AttributeDesc)
     */
    /**
     * @param c the concept the given description belongs to
     * @param desc the attribute description which should be cleaned in all
     *  cases known for concept c
     */
    @Override
    public void clean(final Concept c, final AttributeDesc desc) {
        List<Case> currentCases = cases.get(c);
        if (currentCases != null) {
            for (Case caze : currentCases) {
                caze.clean(desc);
            }
        }
        for (Concept sub : c.getSubConcepts().values()) {
            clean(sub, desc);
        }
    }

    /* (non-Javadoc)
     * @see java.util.Observer#update(java.util.Observable, java.lang.Object)
     */
    /**
     * This observes all cases.
     * On change of any of those cases, the observers of this are notified
     * such that they can react on the changes.
     *
     * @param arg0 observable
     * @param arg1 object
     */
    @Override
    public void update(final Observable arg0, final Object arg1) {
       setChanged();
       notifyObservers();
    }

}
