/*
 * myCBR License 2.0
 *
 * Copyright (c) 2010
 * Thomas Roth-Berghofer, Armin Stahl & German Research Center for Artificial
 * Intelligence DFKI GmbH
 * All rights reserved.
 *
 * myCBR is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Since myCBR uses some modules, you should be aware of their licenses for
 * which you should have received a copy along with this program, too.
 *
 * endOfLic */
package de.dfki.mycbr.core;

import de.dfki.mycbr.core.casebase.Case;
import de.dfki.mycbr.core.casebase.Instance;
import de.dfki.mycbr.core.model.AttributeDesc;
import de.dfki.mycbr.core.model.Concept;

import java.util.Collection;
import java.util.Observer;

/**
 * Methods to be implemented by case bases of the current project.
 *
 * @author myCBR Team
 *
 */
public interface ICaseBase extends Observer {

    /**
     * Returns the name of the case base. The case base is identified by its
     * name
     *
     * @return the unique name of the case base
     */
    String getName();

    /**
     * Sets the name of this case base to name. The name has to be unique, use
     * commands to check uniqueness because case base does not know about the
     * names of other case bases
     *
     * @param name
     *            the unique name of the case base
     * @throws Exception when there is a case base with that name
     */
    void setName(String name) throws Exception;

    /**
     * Returns true, if the case base contains a case with the given name.
     *
     * @param name
     *            the name of the case to be found
     * @return true, if there is a case with this name, false otherwise.
     */
    boolean containsCase(String name);

    /**
     * Adds a new case to this case base which fits the given c and has the
     * specified name.
     *
     * @param i
     *            the conceptDesc for which a case should be added
     * @param id
     *            the unique name for the new case
     * @return the new case
     * @throws Exception when there a case with that name
     */
    Case addCase(Instance i, String id) throws Exception;

    /**
     * Removes the case with the given name. Does nothing if there is none.
     * There can at most be one case with the given name, since cases are
     * identified by their name.
     *
     * @param name
     *            the name of the case to be deleted
     */
    void removeCase(String name);

    /**
     * Removes the case with the given name. Does nothing if there is none.
     * There can at most be one case with the given name, since cases are
     * identified by their name.
     *
     * @param c
     *            the name of the case to be deleted
     */
    void removeCasesOfConcept(Concept c);

    /**
     * Returns all cases known for this case base.
     *
     * @return all cases contained in this case base.
     */
    Collection<Case> getCases();

    /**
     * Returns all cases known for this case base.
     *
     * @param c the concept whose cases should be returned
     * @return all cases contained in this case base.
     */
    Collection<Case> getCases(Concept c);

    /**
     * Returns all cases known for this case base.
     * Not only returns all cases directly associated with
     * the given concept, but also all cases which are valid for
     * the given concept, i.e. all cases known for all ancestors
     * of this concept.
     *
     * @param c the concept whose cases should be returned.
     * @return all cases contained in this case base.
     */
    Collection<Case> getAllCases(Concept c);

    /**
     * Cleans all cases of c and sub concepts of c. Sets the values of cases for
     * attribute description desc to _undefinded_ if the current value does not
     * fit desc any longer
     *
     * @param c the concept whose cases should be cleaned
     * @param desc cases for the given concept might have inconsistent values
     *  with respect to this attribute description
     */
    void clean(Concept c, AttributeDesc desc);
}
