/*
 * myCBR License 2.0
 *
 * Copyright (c) 2010
 * Thomas Roth-Berghofer, Armin Stahl & Deutsches Forschungszentrum f&uuml;r
 * K&uuml;nstliche Intelligenz DFKI GmbH
 * All rights reserved.
 *
 * myCBR is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Since myCBR uses some modules, you should be aware of their licenses for
 * which you should have received a copy along with this program, too.
 *
 * endOfLic */

package de.dfki.mycbr.core.casebase;

import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Observable;

import de.dfki.mycbr.core.ICaseBase;
import de.dfki.mycbr.core.model.AttributeDesc;
import de.dfki.mycbr.core.model.Concept;
import de.dfki.mycbr.core.similarity.Similarity;

/**
 * Cases represent the data in the case base.
 *
 * @author myCBR Team
 *
 */
public final class Case extends Observable {

    /**
     *
     */
    private Instance instance;

    /**
     *
     */
    private String id;

    /**
     *
     */
    private ICaseBase cb;

    private String author;
    private Date date;
    private SimpleDateFormat df = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
    
    /**
     * Initialize this case with the given name and description.
     *
     * @param caseBase the case base this case should belong to
     * @param i
     *            the description of the case
     * @param name
     *            the case's name
     */
    public Case(final ICaseBase caseBase, final Instance i, final String name) 
    	throws Exception {
        this.cb = caseBase;
        this.instance = i;
        if (name == null || name.trim().equals("") || cb.containsCase(name) ) {
        	throw new Exception("Cannot add case. Either name is empty, null or not unique");
        }
        this.id = name; // TODO unique!

        i.addCase(this);
        init();
    }

    /**
     *
     * @return the case base this case belongs to
     */
    public ICaseBase getCB() {
        return cb;
    }

    /**
     *
     * @return the unique identifier of this case
     */
    public String getID() {
        return id;
    }

    /**
     *
     * @param identifier sets the identifier of this case to id
     */
    public void setID(final String identifier) {
        this.id = identifier; // TODO unique
    }

    /**
     * First all attributes get undefined as value. Ensures that there is a
     * value for each attribute description.
     */
    public void init() {
        HashMap<String, Similarity> localSim = new
                                                HashMap<String, Similarity>();
        for (AttributeDesc att : instance.getConcept().getAllAttributeDescs()
                .values()) {
            localSim.put(att.getName(), Similarity.INVALID_SIM);
        }
        
		date = new Date();
        author = instance.getConcept().getProject().getAuthor();
    }

    /**
     *
     * @return the instance this case refers to
     */
    public Instance getInstance() {
        return instance;
    }

    /**
     * @param desc cleans the underlying instance wrt desc
     */
    public void clean(final AttributeDesc desc) {
        instance.clean(desc);
    }

    /**
     * @param attDesc the description whose value should be returned
     * @return the value of this wrt to the given description
     */
    public Attribute getAttForDesc(final AttributeDesc attDesc) {
        return instance.getAttForDesc(attDesc);
    }

    /**
     * @param desc the description of the value to be added to the instance
     * @param a the attribute that should be added to the underlying instance
     * @return true if attribute has been successfully added, false otherwise
     */
    public boolean addAttribute(final AttributeDesc desc, final Attribute a) {
        return instance.addAttribute(desc, a);
    }

    /**
     * @return the concept this case belongs to
     */
    public Concept getConcept() {
        return instance.getConcept();
    }

    /**
     * @param name the name of the description this attribute belongs to
     * @param a the attribute to be added to the underlying instance
     * @return true if attribute has been successfully added, false otherwise
     */
    public boolean addAttribute(final String name, final Attribute a) {
        return instance.addAttribute(name, a);
    }

    /**
     * @param name the name of the description the attribute belongs to
     * @param s the string representation of the value to be added
     * @return true if attribute has been successfully added, false otherwise
     * @throws ParseException if the given string does not fit to the
     *  description given by name
     */
    public boolean addAttribute(final String name, final String s)
                                                        throws ParseException {
        return instance.addAttribute(name, s);
    }

    /**
     * @param desc the description of the attribute to be added
     * @param a the attribute for description desc which should be added
     * @return true if attribute has been successfully added, false otherwise
     * @throws ParseException if the given string does not fit to the
     *  description given by name
     */
    public boolean addAttribute(final AttributeDesc desc, final Object a)
                                                       throws ParseException {
        return instance.addAttribute(desc, a);
    }
    
    ///////////////////////////////////////////////////////////////
    //////////// Accessor methods for additional info /////////////
    ///////////////////////////////////////////////////////////////
    
    /**
     * @since 3.0.0 BETA 0.2
     */
    public String getAuthor() {
    	return author;
    }
    
    /**
     * @since 3.0.0 BETA 0.2
     */
    public void setAuthor(String author) {
    	this.author = author;
    }
    
    /**
     * @since 3.0.0 BETA 0.2
     */
    public Date getDate() {
    	return date;
    }
    
    /**
     * @since 3.0.0 BETA 0.2
     */
    public void setDate(Date date) {
    	this.date = date;
    }
    
    /**
     * @since 3.0.0 BETA 0.2
     */
    public SimpleDateFormat getDateFormat() {
    	return df;
    }
}
