/*
 * myCBR License 2.0
 *
 * Copyright (c) 2009
 * Thomas Roth-Berghofer, Armin Stahl & German Research Center for Artificial Intelligence DFKI GmbH
 * Further contributors: myCBR Team (see http://mycbr-project.net/contact.html for further information 
 * about the myCBR Team). 
 * All rights reserved.
 *
 * myCBR is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Since myCBR uses some modules, you should be aware of their licenses for
 * which you should have received a copy along with this program, too.
 * 
 * endOfLic */
package de.dfki.mycbr.core.similarity;

import java.util.Iterator;
import java.util.Map;
import java.util.Observable;
import java.util.TreeMap;

import de.dfki.mycbr.core.Project;
import de.dfki.mycbr.core.casebase.Attribute;
import de.dfki.mycbr.core.casebase.IntegerAttribute;
import de.dfki.mycbr.core.casebase.MultipleAttribute;
import de.dfki.mycbr.core.casebase.SimpleAttribute;
import de.dfki.mycbr.core.casebase.SpecialAttribute;
import de.dfki.mycbr.core.model.AttributeDesc;
import de.dfki.mycbr.core.model.IntegerDesc;
import de.dfki.mycbr.core.similarity.config.MultipleConfig;

/**
 * Used for calculating similarities of IntegerAttribute objects.
 * This function can also be used for calculating similarity of 
 * ordered symbol attributes. The basis of this function is a simple
 * graph. You can add additional points to this graph. The similarity function
 * is then the interpolation of these points.
 * 
 * @author myCBR Team
 *
 */
public class AdvancedIntegerFct extends Observable implements ISimFct {

	private TreeMap<Double, Similarity> points;
	// the first three points consist of the maximal negative distance, 0 and the 
	// maximal positive distance
	private Similarity minPoint;
	private Similarity zeroPoint;
	private Similarity maxPoint;
	private Project prj;
	private String name;
	private IntegerDesc desc;
	private MultipleConfig mc = MultipleConfig.DEFAULT_CONFIG;
	private Boolean symmetric = true;
	private double max;
	private double min;
	private double diff;
	protected int mode = IntegerFct.DIFFERENCE_MODE;
	
	/**
	 * Initializes this with the given description. This function can
	 * only be applied to attributes with this description.
	 * The points used for interpolating this function are initialized with standard values.
	 * 
	 * @param desc the description of the attributes to be compared
	 */
	public AdvancedIntegerFct(Project prj, IntegerDesc desc, String name) {
		this.prj = prj;
		this.name = name;
		this.desc = desc;
		points = new TreeMap<Double, Similarity>();
		minPoint = Similarity.get(0.00);
		zeroPoint = Similarity.get(1.00);
		maxPoint = Similarity.get(0.00);
		max = desc.getMax();
		min = desc.getMin();
		diff = max - min;
		
		points.put(-diff, minPoint);
		points.put(0d, zeroPoint);
		points.put(diff, maxPoint);
		
	}

	/**
	 * Initializes this with the given description. This function can
	 * only be applied to attributes with this description.
	 * 
	 * @param desc the description of the attributes to be compared
	 */
	AdvancedIntegerFct(Project prj, IntegerDesc desc, Similarity minSim, Similarity zero, Similarity maxSim, String name) {
		this.prj = prj;
		this.desc = desc;
		this.name = name;
		points = new TreeMap<Double, Similarity>();
		minPoint = minSim;
		zeroPoint = zero;
		maxPoint = maxSim;
		max = desc.getMax();
		min = desc.getMin();
		diff = max - min;
		points.put(-diff, minPoint);
		points.put(0d, zeroPoint);
		points.put(diff, maxPoint);
	}
	
	/**
	 * Calculates the similarity of the specified attributes
	 * using the interpolated function. If the attributes do not 
	 * have the given description, an invalid similarity is returned.
	 * 
	 * @return the similarity of the given attributes 
	 *  	or an invalid similarity if an error occurs 
	 * @throws Exception 
	 */
	@Override
	public Similarity calculateSimilarity(Attribute a1, Attribute a2) throws Exception {
		Similarity res = Similarity.INVALID_SIM;
		if (a1 instanceof SpecialAttribute || a2 instanceof SpecialAttribute) {
			res = prj.calculateSpecialSimilarity(a1, a2);
		} else if (a1 instanceof MultipleAttribute<?> && a1 instanceof MultipleAttribute<?>) {
			res = prj
			.calculateMultipleAttributeSimilarity(this,((MultipleAttribute<?>)a1), (MultipleAttribute<?>)a2);
		} else if ((a1 instanceof SimpleAttribute) && (a2 instanceof SimpleAttribute)) {
			SimpleAttribute value1 = (SimpleAttribute)a1;
			SimpleAttribute value2 = (SimpleAttribute)a2;
			
			if (value1.getAttributeDesc()!=desc || value2.getAttributeDesc() != desc) {
				return res;
			} 
			
			 if (value1 instanceof IntegerAttribute && value2 instanceof IntegerAttribute) {
				int q = ((IntegerAttribute)value1).getValue();
				int c = ((IntegerAttribute)value2).getValue();
				
				res = calculateSimilarity(q, c);
			}
		}
		return res;
	}

	public Similarity calculateSimilarity(int q, int c) throws Exception {

		Similarity res;
		double d = mode == IntegerFct.DIFFERENCE_MODE ? c-q : c/q;
		
		// check whether the distance has a special point
		Similarity sim = points.get(d);
		if (sim != null) {
			res = sim;
			return res;
		}
		
		// now it has to be located between two points
		// we find them and interpolate a linear function
		// to get the correct similarity for d
		Map.Entry<Double,Similarity> leftPoint = null;
		Map.Entry<Double,Similarity> rightPoint = null;

		Iterator<Map.Entry<Double,Similarity>> it = points.entrySet().iterator();
		do {
			rightPoint = it.next();
			if (rightPoint.getKey() > d) {
				break;
			}
			leftPoint = rightPoint;
		} while (it.hasNext());
		
		double leftSim = leftPoint.getValue().getValue();
		double rightSim = rightPoint.getValue().getValue();
		
		// now compute the similarity of the given distance
		// by calculating the value of the line given by the two
		// points: y = (y_2-y_1)/(x_2-x_1)*(x-x_1)+y_1
		// where x is our current distance and y 
		// the similarity we are looking for
		double x1 = leftPoint.getKey();
		double x2 = rightPoint.getKey();
		double y1 = leftSim;
		double y2 = rightSim;
		
		double result = (y2-y1)/(x2-x1)*d + (x2*y1 - x1*y2)/(x2-x1);
		return Similarity.get(result);	
	}
	
	/**
	 * Returns the map of additional point used to interpolate this function
	 * Points are pairs of a double value and a similarity
	 * @return the map of additional points
	 */
	public TreeMap<Double, Similarity> getAdditionalPoints() {
		return points;
	}
	
	/**
	 * Adds the specified point to the list of additional points.
	 * @param key Double, the distance describing the first coordinate of the new point
	 * @param value Similarity used as the value for the second coordinate
	 */
	public void addAdditionalPoint(Double key, Similarity value) {
		points.put(key, value);
		setChanged();
		notifyObservers();
	}

	/* (non-Javadoc)
	 * @see de.dfki.mycbr.core.similarity.ISimFct#getDesc()
	 */
	@Override
	public AttributeDesc getDesc() {
		return desc;
	}

	/* (non-Javadoc)
	 * @see de.dfki.mycbr.core.similarity.ISimFct#getMultipleConfig()
	 */
	@Override
	public MultipleConfig getMultipleConfig() {
		return mc;
	}

	/* (non-Javadoc)
	 * @see de.dfki.mycbr.core.similarity.ISimFct#getName()
	 */
	@Override
	public String getName() {
		return name;
	}

	/* (non-Javadoc)
	 * @see de.dfki.mycbr.core.similarity.ISimFct#getProject()
	 */
	@Override
	public Project getProject() {
		return prj;
	}

	/* (non-Javadoc)
	 * @see de.dfki.mycbr.core.similarity.ISimFct#isSymmetric()
	 */
	@Override
	public boolean isSymmetric() {
		return symmetric;
	}

	/* (non-Javadoc)
	 * @see de.dfki.mycbr.core.similarity.ISimFct#setMultipleConfig(de.dfki.mycbr.core.similarity.config.MultipleConfig)
	 */
	@Override
	public void setMultipleConfig(MultipleConfig mc) {
		this.mc = mc;
	}

	/* (non-Javadoc)
	 * @see de.dfki.mycbr.core.similarity.ISimFct#setName(java.lang.String)
	 */
	@Override
	public void setName(String name) {
		this.name = name;
		setChanged();
		notifyObservers();
	}

	/* (non-Javadoc)
	 * @see de.dfki.mycbr.core.similarity.ISimFct#setSymmetric(boolean)
	 */
	@Override
	public void setSymmetric(boolean symmetric) {
		this.symmetric = symmetric; // TODO update
		setChanged();
		notifyObservers();
	}

	/* (non-Javadoc)
	 * @see java.util.Observer#update(java.util.Observable, java.lang.Object)
	 */
	@Override
	public void update(Observable arg0, Object arg1) {
		setChanged();
		notifyObservers();
	}
	
	/**
	 * Sets the mode of this function to mode
	 * @param mode the new mode of this function
	 */
	public void setMode(int mode) {
		this.mode = mode;
		setChanged();
		notifyObservers();
	}

	/**
	 * Returns the mode of this function
	 * @return the mode of this function
	 */
	public int getMode() {
		return mode;
	}

	public double getMin() {
		return min;
	}
	
	public double getMax() {
		return max;
	}
	
	public double getDiff() {
		return diff;
	}
	
	/* (non-Javadoc)
	 * @see de.dfki.mycbr.core.similarity.ISimFct#clone(de.dfki.mycbr.core.model.AttributeDesc, boolean)
	 */
	@SuppressWarnings("unchecked")
	@Override
	public void clone(AttributeDesc descNEW, boolean active) {
		if (descNEW instanceof IntegerDesc && !name.equals(Project.DEFAULT_FCT_NAME)) {
			AdvancedIntegerFct f = ((IntegerDesc)descNEW).addAdvancedIntegerFct(name, active);
			f.mode = this.mode;
			f.points = (TreeMap<Double, Similarity>) this.points.clone();
			f.symmetric = this.symmetric;
			f.zeroPoint = this.zeroPoint;
		}
	}
}
