/*
 * myCBR License 2.0
 *
 * Copyright (c) 2009
 * Thomas Roth-Berghofer, Armin Stahl & German Research Center for Artificial Intelligence DFKI GmbH
 * Further contributors: myCBR Team (see http://mycbr-project.net/contact.html for further information 
 * about the myCBR Team). 
 * All rights reserved.
 *
 * myCBR is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Since myCBR uses some modules, you should be aware of their licenses for
 * which you should have received a copy along with this program, too.
 * 
 * endOfLic */
package de.dfki.mycbr.core.similarity;

import java.util.Observable;

import de.dfki.mycbr.core.Project;
import de.dfki.mycbr.core.casebase.Attribute;
import de.dfki.mycbr.core.casebase.FloatAttribute;
import de.dfki.mycbr.core.casebase.FloatRange;
import de.dfki.mycbr.core.casebase.MultipleAttribute;
import de.dfki.mycbr.core.casebase.SimpleAttribute;
import de.dfki.mycbr.core.casebase.SpecialAttribute;
import de.dfki.mycbr.core.model.AttributeDesc;
import de.dfki.mycbr.core.model.FloatDesc;
import de.dfki.mycbr.core.similarity.config.MultipleConfig;
import de.dfki.mycbr.core.similarity.config.NumberConfig;

/**
 * Calculates the similarity of FloatAttribute objects. There are four modes
 * which can be chosen for the similarity function:
 * <ul>
 * <li>CONSTANT: the similarity is always a constant value</li>
 * <li>STEP AT: for a distance smaller than the specified value the similarity
 * is 0.00. For the other values it is 1.00</li>
 * <li>POLINOMIAL WITH: stretches/compresses the line described by y =
 * x/(max-min) + 1 ( for case < query ) resp. y = - (x/(max-min)) +1 ( for case >
 * query ) by taking y to the power of the given parameter </li>
 * <li>SMOOTH STEP AT: similar to STEP AT but with a smooth step, the
 * function is computed as the logistic function 1/(1 +
 * exp((x-param)*100/(max-min))) ( cp. sigmoid-curve or s-curve)</li>
 * </ul>
 * You have to configure the similarity for two cases:
 * <ul>
 * <li>query < case: referred to as the right part of the function</li>
 * <li>case < query: referred to as the left part of the function</li>
 * </ul>
 * If the function is symmetric, the configurations should be the same. The
 * similarity of a value to itself is assumed to be 1.00 in any case.
 * 
 * @author myCBR Team
 */
public class FloatFct extends Observable implements ISimFct {

	/**
	 * Computes the similarity based on case-query
	 */
	public static final int DIFFERENCE_MODE = 0;

	
	/**
	 * Computes the similarity based on case/query Only allowed in case that 0
	 * is not allowed as query value
	 */
	public static final int QUOTIENT_MODE = 1;

	private String name;
	private final double constantValue = 1.0;
	private final double stepAtValue = 0;
	private final double polynomialWithValue = 1;
	private final double smoothStepAtValue = 0;
	
	private MultipleConfig mc = MultipleConfig.DEFAULT_CONFIG;
	
	/**
	 * query < case
	 */
	private NumberConfig functionTypeR = NumberConfig.CONSTANT;
	private double functionParameterR = constantValue;

	/**
	 * case < query
	 */
	private NumberConfig functionTypeL = NumberConfig.CONSTANT;
	private double functionParameterL = constantValue;

	protected boolean isSymmetric = true;
	protected FloatDesc desc;
	protected FloatRange range;

	protected double max;
	protected double min;
	protected double diff;

	/**
	 * current mode
	 */
	protected int mode = DIFFERENCE_MODE;
	private Project prj;
	
	/**
	 * Constructor should only be called by FunctionContainer. Initializes this
	 * with the given description.
	 * 
	 * @param desc
	 *            the description of the attributes this function can be applied
	 *            to
	 */
	public FloatFct(Project prj, FloatDesc desc, String name) {
		this.prj = prj;
		this.range  = desc.getRange();
		this.desc = desc;

		max = desc.getMax();
		min = desc.getMin();
		diff = max - min;
		this.name = name;
	}

	/**
	 * Returns the similarity of the given attributes. Returns invalid
	 * similarity if an error occurs. The similarity computation depends on the
	 * type of the function and it's parameter. We refer to the right part of
	 * the function if case < query and to the left part otherwise. The possible
	 * computations are:
	 * <ul>
	 * <li>CONSTANT: the similarity is always a constant value</li>
	 * <li>STEP AT: for a distance smaller than the specified value the
	 * similarity is 0.00. For the other values it is 1.00</li>
	 * <li>POLINOMIAL WITH: stretches/compresses the line described by y =
	 * x/(max-min) + 1 ( for case < query ) resp. y = - (x/(max-min)) +1 ( for
	 * case > query ) by taking y to the power of the given parameter </li>
	 * <li>SMOOTH STEP AT: similar to STEP AT but with a smooth step, the
	 * function is computed as the logistic function 1/(1 +
	 * exp((x-param)*100/(max-min))) ( cp. sigmoid-curve or s-curve)</li>
	 * </ul>
	 * 
	 * @return Similarity of the given attributes, invalid similarity if an
	 *         error occurs.
	 * @throws Exception 
	 */
	public Similarity calculateSimilarity(Attribute a1, Attribute a2) throws Exception {
		Similarity result = Similarity.INVALID_SIM;
		if (a1 instanceof SpecialAttribute
				|| a2 instanceof SpecialAttribute) {
		    result = prj
					.calculateSpecialSimilarity(a1, a2);
		} else if (a1 instanceof MultipleAttribute<?> && a2 instanceof MultipleAttribute<?>) {
			result = prj
			.calculateMultipleAttributeSimilarity(this,((MultipleAttribute<?>)a1), (MultipleAttribute<?>)a2);
		} else if (a1 instanceof SimpleAttribute && a2 instanceof SimpleAttribute) {
			SimpleAttribute value1 = (SimpleAttribute)a1;
			SimpleAttribute value2 = (SimpleAttribute)a2;
			
			if ((value1 instanceof FloatAttribute)
					&& (value2 instanceof FloatAttribute)) {
				FloatAttribute att1 = (FloatAttribute) value1;
				FloatAttribute att2 = (FloatAttribute) value2;
	
				// get query and case values
				double q = att1.getValue();
				double c = att2.getValue();
	
				double d = c - q;
				if (mode == QUOTIENT_MODE) {
					if (att1.getValue()==0.0) {
						return Similarity.INVALID_SIM;
					}
					d = c / q;
				}
	
				if (att1.getAttributeDesc() == desc
						&& att2.getAttributeDesc() == desc) {
					if ((d < 0 && mode == DIFFERENCE_MODE)
							|| (d < 1 && mode == QUOTIENT_MODE)) {
	
						switch (functionTypeL) {
						case STEP_AT:
							result = stepAt(d, functionParameterL, true);
							break;
						case POLYNOMIAL_WITH:
							result = polinomialWith(d, functionParameterL, diff);
							break;
						case SMOOTH_STEP_AT:
							result = Similarity.get(1 / (1 + Math
									.exp((-d + functionParameterL) * (100 / diff))));
							break;
						case CONSTANT:
							result = Similarity.get(functionParameterL);
							break;
						}
					} else if ((d > 0 && mode == DIFFERENCE_MODE)
							|| (d > 1 && mode == QUOTIENT_MODE)) {
						switch (functionTypeR) {
							case STEP_AT:
								result =  stepAt(d, functionParameterR, false);
								break;
							case POLYNOMIAL_WITH:
								result =  polinomialWith(d, functionParameterR, -diff);
								break;
							case SMOOTH_STEP_AT:
								result =  Similarity.get(1 / (1 + Math
										.exp((d - functionParameterR) * (100 / diff))));
								break;
							case CONSTANT:
								result = Similarity.get(functionParameterR);
								break;
						}
					} else {
						return Similarity.get(1.00);
					}
				}
			}
		}
		
		return result;
	}

	/**
	 * Returns the similarity for value according to step. If value is less than
	 * step, returns 0.00, else 1.00
	 * 
	 * @param value
	 *            the value for which the similarity value should be returned
	 * @param step
	 *            the step for which similarity switches from 0.00 to 1.00
	 * @return 0.00 if value is less than step, else 1.00
	 */
	private Similarity stepAt(double value, double step, boolean left) {
		if (left) {
			return (value < step) ? Similarity.get(0.00) : Similarity.get(1.00);
		} else {
			return (value > step) ? Similarity.get(0.00) : Similarity.get(1.00);
		}
	}

	/**
	 * Returns the similarity computed by pow(value/diff+1, exponent).
	 * 
	 * @param value
	 *            the value for which the similarity value should be returned
	 * @param exponent
	 *            double value describing how the line given by x/diff+1 should
	 *            be stretched/compressed
	 * @return pow(value/diff+1, exponent)
	 */
	private Similarity polinomialWith(double value, double exponent, double diff) {
		return Similarity.get(Math.pow(value / diff + 1, exponent));
	}

	/**
	 * Calculates the similarity of the given Floats. Calls
	 * {@link #calculateSimilarity(Attribute, Attribute)}. Be aware that
	 * <code>calculateSimilarity(value1, values2)</code> might differ from
	 * <code>calculateSimilarity(value2, values1)</code>. When calculating
	 * similarities, the first attribute is always referred to as the query
	 * value, the second parameter as the case value.
	 * 
	 * @param value1
	 *            the first attribute to be compared
	 * @param value2
	 *            the second attribute to be compared
	 * @return similarity of the given attributes
	 * @throws Exception 
	 */
	public Similarity calculateSimilarity(Float value1, Float value2) throws Exception {
		return calculateSimilarity((SimpleAttribute)range.getAttribute(value1),(SimpleAttribute) range
				.getAttribute(value2));
	}

	/**
	 * Returns whether this function is symmetric or not.
	 * 
	 * @return true, if this function is symmetric, false otherwise
	 */
	public boolean isSymmetric() {
		return isSymmetric;
	}

	/**
	 * Sets this function's symmetry mode.
	 * 
	 * @param symmetric 
	 *            true, if this function is symmetric, false otherwise
	 */
	public void setSymmetric(boolean symmetric) {
		isSymmetric = symmetric;

		this.setChanged();
		notifyObservers();
	}

	/**
	 * Sets the function's parameter for query &lt; case to value. The function
	 * parameter depends on the type of the function. The parameter specifies:
	 * <ul>
	 * <li>CONSTANT: the constant similarity value</li>
	 * <li>STEP AT: the distance value where the step should be (every distance
	 * greater or equal than this value has similarity 1.00, else 0.00) </li>
	 * <li>POLINOMIAL WITH: the constant with which the function gets
	 * compressed or stretched</li>
	 * <li>SMOOTH STEP AT: the distance value where the smooth step should be</li>
	 * </ul>
	 * If the parameter for &quot;smooth step&quot; or &quot;step at&quot; does
	 * not fit the range given by max/min, the parameter will not be updated. If
	 * the parameter for &quot;constant&quot; is not within the range (0.0,1.0),
	 * the parameter will not be updated. Or if the parameter for
	 * &quot;polynomial with&quot; is smaller than 0 the parameter will not be
	 * updated. Be aware to set the function type before setting the parameter,
	 * because the parameter depends on the current function type.
	 * 
	 * @param value
	 *            the function parameter for query &lt; case
	 * @return true, if the parameter was successfully updated, false otherwise
	 */
	public boolean setFunctionParameterR(double value) {

		boolean updated = false;

		if (functionTypeR == NumberConfig.SMOOTH_STEP_AT
				|| functionTypeR == NumberConfig.STEP_AT) {
			if (mode == DIFFERENCE_MODE && value > 0 && value <= diff) {
				functionParameterR = value;
				updated = true;
			} else if (mode == QUOTIENT_MODE && value > 1 && value <= max / min) {
				functionParameterR = value;
				updated = true;
			}
		} else if (functionTypeR == NumberConfig.CONSTANT) {
			if (value >= 0 && value <= 1) {
				functionParameterR = value;
				updated = true;
			}
		} else if (functionTypeR == NumberConfig.POLYNOMIAL_WITH) {
			if (value >= 0) {
				functionParameterR = value;
				updated = true;
			}
		}

		if (updated && isSymmetric()) {
			if (functionTypeL == NumberConfig.SMOOTH_STEP_AT
					|| functionTypeL == NumberConfig.STEP_AT) {
				if (mode == DIFFERENCE_MODE) {
					functionParameterL = -value;
				} else if (mode == QUOTIENT_MODE) {
					functionParameterL = 1 / value;
				}
			} else {
				functionParameterL = value;
			}
			this.setChanged();
			notifyObservers();
		}
		
		return updated;
	}

	/**
	 * Sets the function's parameter for query &lt; case to value. The function
	 * parameter depends on the type of the function. The parameter specifies:
	 * <ul>
	 * <li>CONSTANT: the constant similarity value</li>
	 * <li>STEP AT: the distance value where the step should be (every distance
	 * greater or equal than this value has similarity 1.00, else 0.00) </li>
	 * <li>POLINOMIAL WITH: the constant with which the function gets
	 * compressed or stretched</li>
	 * <li>SMOOTH STEP AT: the distance value where the smooth step should be</li>
	 * </ul>
	 * If the parameter for &quot;smooth step&quot; or &quot;step at&quot; does
	 * not fit the range given by max/min, the parameter will not be updated. If
	 * the parameter for &quot;constant&quot; is not within the range (0.0,1.0),
	 * the parameter will not be updated. Or if the parameter for
	 * &quot;polynomial with&quot; is smaller than 0 the parameter will not be
	 * updated. Be aware to set the function type before setting the parameter,
	 * because the parameter depends on the current function type.
	 * 
	 * @param value
	 *            the function parameter for query &lt; case
	 * @return true, if the parameter was successfully updated, false otherwise
	 */
	public boolean setFunctionParameterL(double value) {

		boolean updated = false;

		if (functionTypeL == NumberConfig.SMOOTH_STEP_AT
				|| functionTypeL == NumberConfig.STEP_AT) {
			if (value > 0) {
				value = -value;
			}
			if (mode == DIFFERENCE_MODE && value >= -diff) {
				
				functionParameterL = value;
				updated = true;
			} else if (mode == QUOTIENT_MODE && value < 1 && value >= min / max) {
				functionParameterL = value;
				updated = true;
			}
		} else if (functionTypeL == NumberConfig.CONSTANT) {
			if (value >= 0 && value <= 1) {
				functionParameterL = value;
				updated = true;
			}
		} else if (functionTypeL == NumberConfig.POLYNOMIAL_WITH) {
			if (value >= 0) {
				functionParameterL = value;
				updated = true;
			}
		}

		if (updated && isSymmetric()) {
			if (functionTypeR == NumberConfig.SMOOTH_STEP_AT
					|| functionTypeR == NumberConfig.STEP_AT) {
				if (mode == DIFFERENCE_MODE) {
					functionParameterR = -value;
				} else if (mode == QUOTIENT_MODE) {
					functionParameterR = 1 / value;
				}
			} else {
				functionParameterR = value;
			}

			this.setChanged();
			notifyObservers();
		}

		return updated;
	}

	/**
	 * Gets the function's parameter for query &lt; case to value. The function
	 * parameter depends on the type of the function. The parameter specifies:
	 * <ul>
	 * <li>CONSTANT: the constant similarity value</li>
	 * <li>STEP AT: the distance value where the step should be (every distance
	 * greater or equal than this value has similarity 1.00, else 0.00) </li>
	 * <li>POLINOMIAL WITH: the constant with which the function gets
	 * compressed or stretched</li>
	 * <li>SMOOTH STEP AT: the distance value where the smooth step should be</li>
	 * </ul>
	 * 
	 * @return the function parameter for query &lt; case
	 */
	public double getFunctionParameterR() {
		return functionParameterR;
	}

	/**
	 * Gets the function's parameter for case &lt; query to value. The function
	 * parameter depends on the type of the function. The parameter specifies:
	 * <ul>
	 * <li>CONSTANT: the constant similarity value</li>
	 * <li>STEP AT: the distance value where the step should be (every distance
	 * greater or equal than this value has similarity 1.00, else 0.00) </li>
	 * <li>POLINOMIAL WITH: the constant with which the function gets
	 * compressed or stretched</li>
	 * <li>SMOOTH STEP AT: the distance value where the smooth step should be</li>
	 * </ul>
	 * 
	 * @return the function parameter for case &lt; query
	 */
	public double getFunctionParameterL() {
		return functionParameterL;
	}

	/**
	 * Sets the type of the function for query &lt; case.
	 * Updates function parameter if necessary.
	 * Be aware to set the function type before setting the parameter, because the parameter depends
	 * on the current function type.
	 * 
	 * @param type
	 *            the type of the function
	 */
	public void setFunctionTypeR(NumberConfig type) {

		if (functionTypeR == NumberConfig.POLYNOMIAL_WITH
				|| functionTypeR == NumberConfig.CONSTANT) {
			if (type == NumberConfig.STEP_AT) {
				functionParameterR = stepAtValue;
			} else if (type == NumberConfig.SMOOTH_STEP_AT) {
				functionParameterR = smoothStepAtValue;
			}
		} else if (functionTypeR == NumberConfig.STEP_AT
				|| functionTypeR == NumberConfig.SMOOTH_STEP_AT) {
			if (type == NumberConfig.POLYNOMIAL_WITH) {
				functionParameterR = polynomialWithValue;
			} else if (type == NumberConfig.CONSTANT) {
				functionParameterR = constantValue;
			}
		}

		this.functionTypeR = type;
		if (isSymmetric()) {
			this.functionTypeL = functionTypeR;
			if (functionTypeL == NumberConfig.SMOOTH_STEP_AT
					|| functionTypeL == NumberConfig.STEP_AT) {
				if (mode == DIFFERENCE_MODE) {
					functionParameterL = -functionParameterR;
				} else if (mode == QUOTIENT_MODE) {
					functionParameterL = 1 / functionParameterR;
				}
			} else {
				functionParameterL = functionParameterR;
			}
		}

		this.setChanged();
		notifyObservers();
	}

	/**
	 * Sets the type of the function for case &lt; query.
	 * Updates function parameter if necessary.
	 * Be aware to set the function type before setting the parameter, because the parameter depends
	 * on the current function type.
	 * 
	 * @param type
	 *            the type of the function
	 */
	public void setFunctionTypeL(NumberConfig type) {
		if (functionTypeL == NumberConfig.POLYNOMIAL_WITH
				|| functionTypeL == NumberConfig.CONSTANT) {
			if (type == NumberConfig.STEP_AT) {
				functionParameterL = stepAtValue;
			} else if (type == NumberConfig.SMOOTH_STEP_AT) {
				functionParameterL = smoothStepAtValue;
			}
		} else if (functionTypeL == NumberConfig.STEP_AT
				|| functionTypeL == NumberConfig.SMOOTH_STEP_AT) {
			if (type == NumberConfig.POLYNOMIAL_WITH) {
				functionParameterL = polynomialWithValue;
			} else if (type == NumberConfig.CONSTANT) {
				functionParameterL = constantValue;
			}
		}

		this.functionTypeL = type;
		if (isSymmetric()) {
			this.functionTypeR = functionTypeL;
			if (functionTypeR == NumberConfig.SMOOTH_STEP_AT
					|| functionTypeR == NumberConfig.STEP_AT) {
				if (mode == DIFFERENCE_MODE) {
					functionParameterR = -functionParameterL;
				} else if (mode == QUOTIENT_MODE) {
					functionParameterR = 1 / functionParameterL;
				}
			} else {
				functionParameterR = functionParameterL;
			}
		}

		this.setChanged();
		notifyObservers();
	}

	/* (non-Javadoc)
	 * @see de.dfki.mycbr.core.similarity.ISimFct#getName()
	 */
	@Override
	/**
	 * Returns the name of this function
	 * @return the name of this function
	 */
	public String getName() {
		return name;
	}

	/* (non-Javadoc)
	 * @see de.dfki.mycbr.core.similarity.ISimFct#setName(java.lang.String)
	 */
	@Override
	/**
	 * Sets the name of this function to name.
	 * Does nothing if there is another function with this name.
	 * @param name the name of this function
	 */
	public void setName(String name) {
		if (desc.getFct(name) == null) {
			desc.renameFct(this.name, name);
			this.name = name;
			setChanged();
			notifyObservers();
		}
	}

	/* (non-Javadoc)
	 * @see de.dfki.mycbr.core.similarity.ISimFct#getDescription()
	 */
	@Override
	/**
	 * Returns the description of the attributes which can 
	 * be compared using this function.
	 * @return description this function belongs to
	 */
	public AttributeDesc getDesc() {
		return desc;
	}
	
	/**
	 * Sets the mode of this function to mode
	 * @param mode the new mode of this function
	 */
	public void setMode(int mode) {
		this.mode = mode;
	}

	/**
	 * Returns the mode of this function
	 * @return the mode of this function
	 */
	public int getMode() {
		return mode;
	}

	/* (non-Javadoc)
	 * @see de.dfki.mycbr.core.similarity.ISimFct#getProject()
	 */
	@Override
	public Project getProject() {
		return prj;
	}

	/* (non-Javadoc)
	 * @see de.dfki.mycbr.core.similarity.ISimFct#getMultipleConfig()
	 */
	@Override
	public MultipleConfig getMultipleConfig() {
		return mc;
	}

	/* (non-Javadoc)
	 * @see de.dfki.mycbr.core.similarity.ISimFct#setMultipleConfig()
	 */
	@Override
	public void setMultipleConfig(MultipleConfig mc) {
		this.mc = mc;
		setChanged();
		notifyObservers();
	}

	/* (non-Javadoc)
	 * @see java.util.Observer#update(java.util.Observable, java.lang.Object)
	 */
	@Override
	public void update(Observable o, Object arg) {
		setChanged();
		notifyObservers();
	}
	
	public NumberConfig getFunctionTypeL() {
		return functionTypeL;
	}
	
	public NumberConfig getFunctionTypeR() {
		return functionTypeR;
	}

	public double getMin() {
		return min;
	}
	
	public double getMax() {
		return max;
	}
	
	public double getDiff() {
		return diff;
	}
	
	/* (non-Javadoc)
	 * @see de.dfki.mycbr.core.similarity.ISimFct#clone(de.dfki.mycbr.core.model.AttributeDesc, boolean)
	 */
	@Override
	public void clone(AttributeDesc descNEW, boolean active) {
		if (descNEW instanceof FloatDesc && !name.equals(Project.DEFAULT_FCT_NAME)) {
			FloatFct f = ((FloatDesc)descNEW).addFloatFct(name, active);
			f.functionParameterL = this.functionParameterL;
			f.functionParameterR = this.functionParameterR;
			f.functionTypeL = this.functionTypeL;
			f.functionTypeR = this.functionTypeR;
			f.isSymmetric = this.isSymmetric;
			f.mc = this.mc;
			f.mode = this.mode;
		}
	}
	
}
