/*
 * User.cpp
 *
 *  Created on: Apr 22, 2011
 *      Author: johnsverre
 */

#include "User.h"
#include "UserPoints.h"
#include <iostream>
#include <cmath>
#include <string>
#include <sstream>
#include <XnCppWrapper.h>

using namespace std;
using namespace xn;

User::User(){};

User::User(XnUserID user) {
	this->xnid = user;
	this->angleDeg = true;
	this->maxUserPoints = 200;
	this->maxFeatureSets = 200;
	this->setIntention(false);
}

User::~User() {
	// TODO Auto-generated destructor stub
}

XnUserID User::getUserID() const{
	return this->xnid;
}

bool User::validUserPoints(UserPoints up) const{
	if(up.isValid()) return true;
	return false;
}

void User::addUserPoints(){
	extern UserGenerator g_UserGenerator;

	if(!g_UserGenerator.GetSkeletonCap().IsCalibrated(this->xnid)) return;
	if(!g_UserGenerator.GetSkeletonCap().IsTracking(this->xnid)) return;

	UserPoints up(g_UserGenerator, this->xnid);

	if (validUserPoints(up)){
		this->userPoints.push_back(up);
		// remove oldest element if max limit is reached
		if(countUserPoints() > maxUserPoints){
			userPoints.erase(userPoints.begin());
		}
	}
}

int User::countUserPoints() const{
	return this->userPoints.size();
}

UserPoints User::getLatestUserPoints() const{
	if(!this->userPoints.empty()){
		return this->userPoints.back();
	}
	return UserPoints();
}

/*
 * Calculate distance from sensor to representative point on user
 * */
double User::calcDistance(UserPoints up) const{
	return up.calcSensorDistance();
}


/*
 * Calculate the alignment of the users body combining shoulder and hip angles
 * */
double User::calcBodyAlignment(UserPoints up) const{
	double shoulderAngle = up.calcShoulderAngle();
	double hipAngle = up.calcHipAngle();

	return (shoulderAngle + hipAngle) / 2.0;
}


/*
 * calculate angle between heading and straight line to sensor
 * based on two different userpoint collections
 * */
double User::calcHeading(UserPoints up1, UserPoints up2) const{

	// get representative points from point collection
	XnPoint3D p0 = up1.getTorso();
	XnPoint3D p1 = up2.getTorso();

	// define vectors
	double vS[] = {-p0.X, -p0.Z};
	double vH[] = {p1.X-p0.X, p1.Z-p0.Z};

	double lH = sqrt(vH[0]*vH[0] + vH[1]*vH[1]);
	if(lH < 20) return -1; // length of vector is too short

	return calcVectorAngle(vS, vH);
}

double User::calcSpeed(UserPoints up1, UserPoints up2) const{
	int time = (up2.getTimestamp() - up1.getTimestamp())/1000; // time in milliseconds

	// get representative points from point collection
	XnPoint3D p0 = up1.getTorso();
	XnPoint3D p1 = up2.getTorso();

	double v[] = {p1.X-p0.X, p1.Z-p0.Z};
	double vL = sqrt(v[0]*v[0] + v[1]*v[1]); // length in mm

	// calculate speed in m/s (mm/ms)
	double speed = vL / time;
	return speed;
}

double User::calcAcceleration(UserPoints up1, UserPoints up2, UserPoints up3) const{
	int time = (up3.getTimestamp() - up2.getTimestamp())/1000;

	double v0 = calcSpeed(up1, up2);
	double v = calcSpeed(up2, up3);

	double acc = (v-v0)/time;

	return acc;
}

double User::calcAngularVelocity(UserPoints up1, UserPoints up2, UserPoints up3) const{
	int time = (up3.getTimestamp() - up2.getTimestamp())/1000;
	double a0 = calcHeading(up1, up2);
	double a = calcHeading(up2, up3);
	
	double av = (a-a0)/time;

	return av;
}

void User::calcFeatures(){
	int size = this->userPoints.size();
	UserPoints up1, up2, up3;
	double distance, bodyAlignment, speed, heading, acceleration, angularVelocity;
	FeatureSet fs;
	bool addFeatureSet = false;

	if(size > 0){ 	// calc the ones that require only one set of points
		up1 = this->userPoints.at(size-1);

		// check if new userpoints
		if(up1.getTimestamp() <= this->getLatestFeatureSet().timestamp){
			return;
		}

		distance = calcDistance(up1);
		bodyAlignment = calcBodyAlignment(up1);
		addFeatureSet = true;

		fs.distance = distance;
		fs.body_alignment = bodyAlignment;
	}
	if(size > 9){
		up2 = this->userPoints.at(size-10);
		speed = calcSpeed(up2, up1);
		heading = calcHeading(up2, up1);

		fs.speed = speed;
		fs.heading = heading;
	}
	if(size > 19){
		up3 = this->userPoints.at(size-20);
		acceleration = calcAcceleration(up3, up2, up1);
		angularVelocity = calcAngularVelocity(up3, up2, up1);

		fs.acceleration = acceleration;
		fs.angular_velocity = angularVelocity;
	}

	if(addFeatureSet){
		fs.timestamp = up1.getTimestamp();
		featureStack.push(fs);
		if(countFeatureSets() > maxFeatureSets){
			featureStack.pop();
		}

	}
}


FeatureSet User::getLatestFeatureSet() const{
	if(!this->featureStack.empty()){
		return this->featureStack.back();
	}
	return FeatureSet();
}

int User::countFeatureSets() const{
	return this->featureStack.size();
}

/*
 * calculate angle between the vectors v1 and v2
 * */
double User::calcVectorAngle(double *v1, double *v2) const{

	// vector length
	double lV1 = sqrt(v1[0]*v1[0] + v1[1]*v1[1]);
	double lV2 = sqrt(v2[0]*v2[0] + v2[1]*v2[1]);

	if(lV1*lV2 == 0) return 0;

	double angle = acos((v1[0] * v2[0] + v1[1]*v2[1])/(lV1*lV2));

	if(angleDeg){
		return 180*angle/3.141592654;
	}else{
		return angle;
	}

}


int User::getPositiveIndications() const{
	return positiveIndications;
}

void User::setPositiveIndications(int nPositive){
	if(nPositive > 0)
		this->positiveIndications = nPositive;
	else
		this->positiveIndications = 0;
}

bool User::getIntention() const{
	return this->intention;
}

void User::setIntention(bool intention) {
	this->intention = intention;
}

string User::toString() const{
	stringstream ss;
	ss << "User: " << this->xnid << endl;
	ss << "\tUserPoints: " << this->userPoints.size() <<  " item(s)" << endl;
	ss << "\tFeatureStack: " << this->featureStack.size() <<  " item(s)" << endl;
	return ss.str();
}
