/*
 * UserPoints.cpp
 *
 *  Created on: Apr 18, 2011
 *      Author: johnsverre
 */

#include "UserPoints.h"
#include <XnCppWrapper.h>
#include <cmath>
#include <iostream>
#include <string>
#include <sstream>

using namespace std;

UserPoints::UserPoints(){
	XnPoint3D blank;
	blank.X = 0;
	blank.Y = 0;
	blank.Z = 0;
	this->joints[0] = blank;
	this->joints[1] = blank;
	this->joints[2] = blank;
	this->joints[3] = blank;
	this->joints[4] = blank;
	this->timestamp = 0;
}

UserPoints::UserPoints(UserGenerator &userGenerator, XnUserID user) {

	XnSkeletonJointPosition joint1, joint2, joint3, joint4, joint5;
	userGenerator.GetSkeletonCap().GetSkeletonJointPosition(user, XN_SKEL_LEFT_SHOULDER, joint1);
	userGenerator.GetSkeletonCap().GetSkeletonJointPosition(user, XN_SKEL_RIGHT_SHOULDER, joint2);
	userGenerator.GetSkeletonCap().GetSkeletonJointPosition(user, XN_SKEL_LEFT_HIP, joint3);
	userGenerator.GetSkeletonCap().GetSkeletonJointPosition(user, XN_SKEL_RIGHT_HIP, joint4);
	joint5.position = calcMiddlePoint(joint1.position, joint2.position);
	joint5.fConfidence = 1;
	// temporary fix to reduce number of crashes by NITE, use calcualted middlepoint instead of torso
	//userGenerator.GetSkeletonCap().GetSkeletonJointPosition(user, XN_SKEL_TORSO, joint5);

	float t = 0.5f;
	if(joint1.fConfidence < t || joint2.fConfidence < t || joint3.fConfidence < t || joint4.fConfidence < t || joint5.fConfidence < t){
		XnPoint3D blank;
		blank.X = 0;
		blank.Y = 0;
		blank.Z = 0;
		this->joints[0] = blank;
		this->joints[1] = blank;
		this->joints[2] = blank;
		this->joints[3] = blank;
		this->joints[4] = blank;
		this->timestamp = 0;
	}else{

		this->timestamp = userGenerator.GetTimestamp();

		this->door.X = 0;
		this->door.Y = 0;
		this->door.Z = 0;

		this->joints[0] = joint1.position;
		this->joints[1] = joint2.position;
		this->joints[2] = joint3.position;
		this->joints[3] = joint4.position;
		this->joints[4] = joint5.position;

		this->angleDeg = true;
	}
}

UserPoints::~UserPoints() {
	// TODO Auto-generated destructor stub
}

/*
 * Calculates the angle between the line through points p1, p2 and the tangent line
 * to the sensor radius
 *
 * */
float UserPoints::calcPointAngle(XnPoint3D p1, XnPoint3D p2) const{

	XnPoint3D middle;
	middle.X = ((p1.X-p2.X)/2)+p2.X;
	middle.Z = ((p1.Z-p2.Z)/2)+p2.Z;

	// find slope of tangent line
	float aT = -(middle.X/middle.Z);

	// find intersect with Y-axis
	float bT = middle.Z - aT*middle.X;

	// define vectors
	float vt[] = {0-middle.X, bT-middle.Z};


	float vs[2];

	// use L/R-point according to quadrant
	if(middle.X < 0){
		vs[0] = p2.X-middle.X;
		vs[1] = p2.Z-middle.Z;
	}else{
		vs[0] = p1.X-middle.X;
		vs[1] = p1.Z-middle.Z;
	}

	// vector length
	float lT = sqrt(vt[0]*vt[0] + vt[1]*vt[1]);
	float lS = sqrt(vs[0]*vs[0] + vs[1]*vs[1]);

	float angle = acos((vt[0]*vs[0] + vt[1]*vs[1])/(lT*lS));

	//if(angle>(M_PI/2)) angle = M_PI - angle;

	if(angleDeg){
		return 180*angle/3.141592654;
	}else{
		return angle;
	}
}

float UserPoints::calcShoulderAngle() const{
	return calcPointAngle(joints[0], joints[1]);
}

float UserPoints::calcHipAngle() const{
	return calcPointAngle(joints[2], joints[3]);
}

int UserPoints::calcSensorDistance() const{
	return (int) sqrt(joints[4].Z*joints[4].Z+joints[4].X*joints[4].X);
}

XnPoint3D UserPoints::calcMiddlePoint(XnPoint3D &p1, XnPoint3D &p2) const{
	XnPoint3D middle;

	float dX = p1.X - p2.X;
	float dY = p1.Y - p2.Y;
	float dZ = p1.Z - p2.Z;

	middle.X = dX/2 +  p2.X;
	middle.Y = dY/2 +  p2.Y;
	middle.Z = dZ/2 +  p2.Z;

	return middle;
}

XnPoint3D UserPoints::getLShoulder(){
	return joints[0];
}
XnPoint3D UserPoints::getRShoulder(){
	return joints[1];
}
XnPoint3D UserPoints::getLHip(){
	return joints[2];
}
XnPoint3D UserPoints::getRHip(){
	return joints[3];
}
XnPoint3D UserPoints::getTorso(){
	return joints[4];
}

XnPoint3D* UserPoints::getJoints(){
	return joints;
}

XnUInt64 UserPoints::getTimestamp(){
	return timestamp;
}

void UserPoints::setLShoulder(XnPoint3D ls){
	this->joints[0] = ls;
}

void UserPoints::setRShoulder(XnPoint3D rs){
	this->joints[1] = rs;
}

void UserPoints::setLHip(XnPoint3D lh){
	this->joints[2] = lh;
}

void UserPoints::setRHip(XnPoint3D rh){
	this->joints[3] = rh;
}

void UserPoints::setTorso(XnPoint3D t){
	this->joints[4] = t;
}

void UserPoints::setJoints(XnPoint3D *jointArr){
	this->joints[0] = jointArr[0];
	this->joints[1] = jointArr[1];
	this->joints[2] = jointArr[2];
	this->joints[3] = jointArr[3];
	this->joints[4] = jointArr[4];
}


bool UserPoints::isValid() const{
	// dummy point
	if(this->timestamp == 0) return false;

	for(int i=0;i<5;i++){
		XnFloat X = joints[i].X;
		XnFloat Y = joints[i].Y;
		XnFloat Z = joints[i].Z;

		if(X==0 && Y==0 && Z==0) return false;

		if(Z <= 50 || Z >= 15000){
			//cout << "Z out of range: " << Z << endl;
			return false;
		}
		if(Y >= 15000 || Y <= -15000){
			//cout << "Y out of range: " << Y << endl;
			return false;
		}
		if(X >= 15000 || X <= -15000){
			//cout << "X out of range: " << X << endl;
			return false;
		}
	}
	return true;
}

string UserPoints::toString() const{
	stringstream ss;
	ss << "UserPoints (" << timestamp << "): " << endl;
	ss << "\tJoint0: [" << joints[0].X << "," << joints[0].Y << "," << joints[0].Z << "]" << endl;
	ss << "\tJoint1: [" << joints[1].X << "," << joints[1].Y << "," << joints[1].Z << "]" << endl;
	ss << "\tJoint2: [" << joints[2].X << "," << joints[2].Y << "," << joints[2].Z << "]" << endl;
	ss << "\tJoint3: [" << joints[3].X << "," << joints[3].Y << "," << joints[3].Z << "]" << endl;
	ss << "\tJoint4: [" << joints[4].X << "," << joints[4].Y << "," << joints[4].Z << "]" << endl;

	return ss.str();
}
