//---------------------------------------------------------------------------
// Includes
//---------------------------------------------------------------------------
#include <XnCppWrapper.h>
#include <iostream>
#include <iomanip>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <sstream>
#include "Server.h"
#include "UserPoints.h"
#include "UserHandler.h"
#include "User.h"

#include <cmath>
#include <fstream>

using namespace std;
using namespace cv;
using namespace xn;



//---------------------------------------------------------------------------
// Globals
//---------------------------------------------------------------------------
xn::Context g_Context;
xn::DepthGenerator g_DepthGenerator;
xn::UserGenerator g_UserGenerator;
UserHandler g_UserHandler;

XnBool g_bNeedPose = FALSE;
XnChar g_strPose[20] = "";
XnBool g_bDrawBackground = TRUE;
XnBool g_bDrawPixels = TRUE;
XnBool g_bDrawSkeleton = TRUE;
XnBool g_bPrintID = TRUE;
XnBool g_bPrintState = TRUE;

XnBool g_bPause = false;
XnBool g_bRecord = false;

XnBool g_bQuit = false;

//---------------------------------------
// Defines
//-------------------------------------------

#define SAMPLE_XML_PATH "Sample-User.xml"
#define _USE_MATH_DEFINES
#define CHECK_RC(nRetVal, what)										\
	if (nRetVal != XN_STATUS_OK)									\
	{																\
		printf("%s failed: %s\n", what, xnGetStatusString(nRetVal));\
		return nRetVal;												\
	}

#define XN_CALIBRATION_FILE_NAME "calibration.bin"

//-----------------------------
// Prototypes
//------------------------------

void LoadCalibration();
void SaveCalibration();
void PrintRC(XnStatus rc, const char *what);

//---------------------------------------------------------------------------
// Code
//---------------------------------------------------------------------------




void CleanupExit()
{
	g_Context.Shutdown();

	exit (0);
}

// Callback: New user was detected
void XN_CALLBACK_TYPE User_NewUser(xn::UserGenerator& generator, XnUserID nId, void* pCookie)
{
	printf("New User %d\n", nId);
	// New user found

	g_UserHandler.addUser(User(nId));

	LoadCalibration();

	if(g_UserGenerator.GetSkeletonCap().IsCalibrated(nId)) return;

	if (g_bNeedPose)
	{
		g_UserGenerator.GetPoseDetectionCap().StartPoseDetection(g_strPose, nId);
	}
	else
	{
		g_UserGenerator.GetSkeletonCap().RequestCalibration(nId, TRUE);
	}
}


// Callback: An existing user has exited
void XN_CALLBACK_TYPE User_Exit(xn::UserGenerator& generator, XnUserID nId, void* pCookie)
{
	printf("Lost user %d\n", nId);
	g_UserHandler.removeUser(nId);
}

// Callback: An existing user has re-entered
void XN_CALLBACK_TYPE User_ReEnter(xn::UserGenerator& generator, XnUserID nId, void* pCookie)
{
	printf("User %d has re-entered\n", nId);
	g_UserHandler.addUser(User(nId));

	LoadCalibration();

	if(g_UserGenerator.GetSkeletonCap().IsCalibrated(nId)) return;

	if (g_bNeedPose)
	{
		g_UserGenerator.GetPoseDetectionCap().StartPoseDetection(g_strPose, nId);
	}
	else
	{
		g_UserGenerator.GetSkeletonCap().RequestCalibration(nId, TRUE);
	}
}
// Callback: Detected a pose
void XN_CALLBACK_TYPE UserPose_PoseDetected(xn::PoseDetectionCapability& capability, const XnChar* strPose, XnUserID nId, void* pCookie)
{
	printf("Pose %s detected for user %d\n", strPose, nId);
	g_UserGenerator.GetPoseDetectionCap().StopPoseDetection(nId);
	g_UserGenerator.GetSkeletonCap().RequestCalibration(nId, TRUE);
}
// Callback: Started calibration
void XN_CALLBACK_TYPE UserCalibration_CalibrationStart(xn::SkeletonCapability& capability, XnUserID nId, void* pCookie)
{
	printf("Calibration started for user %d\n", nId);
}
// Callback: Finished calibration
void XN_CALLBACK_TYPE UserCalibration_CalibrationEnd(xn::SkeletonCapability& capability, XnUserID nId, XnBool bSuccess, void* pCookie)
{
	if (bSuccess)
	{
		// Calibration succeeded
		printf("Calibration complete, start tracking user %d\n", nId);
		g_UserGenerator.GetSkeletonCap().StartTracking(nId);
	}
	else
	{
		// Calibration failed
		printf("Calibration failed for user %d\n", nId);
		if (g_bNeedPose)
		{
			g_UserGenerator.GetPoseDetectionCap().StartPoseDetection(g_strPose, nId);
		}
		else
		{
			g_UserGenerator.GetSkeletonCap().RequestCalibration(nId, TRUE);
		}
	}
}


// Save calibration to file
void SaveCalibration()
{
	XnUserID aUserIDs[20] = {0};
	XnUInt16 nUsers = 20;
	g_UserGenerator.GetUsers(aUserIDs, nUsers);

	cout << "There are " << nUsers << " users." << endl;

	for (int i = 0; i < nUsers; ++i)
	{
		cout << "looping through users " << i << endl;
		// Find a user who is already calibrated
		if (g_UserGenerator.GetSkeletonCap().IsCalibrated(aUserIDs[i]))
		{
			// Save user's calibration to file
			cout << "found calibrated user: " << aUserIDs[i] << ", saving to file ..." << endl;
			XnStatus rc = g_UserGenerator.GetSkeletonCap().SaveCalibrationDataToFile(aUserIDs[i], XN_CALIBRATION_FILE_NAME);
			PrintRC(rc, "save to file failed");
			break;
		}
	}
}
// Load calibration from file
void LoadCalibration()
{
	XnUserID aUserIDs[20] = {0};
	XnUInt16 nUsers = 20;
	g_UserGenerator.GetUsers(aUserIDs, nUsers);
	for (int i = 0; i < nUsers; ++i)
	{
		// Find a user who isn't calibrated or currently in pose
		if (g_UserGenerator.GetSkeletonCap().IsCalibrated(aUserIDs[i])) continue;
		if (g_UserGenerator.GetSkeletonCap().IsCalibrating(aUserIDs[i])) continue;

		// Load user's calibration from file
		XnStatus rc = g_UserGenerator.GetSkeletonCap().LoadCalibrationDataFromFile(aUserIDs[i], XN_CALIBRATION_FILE_NAME);
		if (rc == XN_STATUS_OK)
		{
			// Make sure state is coherent
			g_UserGenerator.GetPoseDetectionCap().StopPoseDetection(aUserIDs[i]);
			g_UserGenerator.GetSkeletonCap().StartTracking(aUserIDs[i]);
		}
		break;
	}
}

Point Xn2CVPoint(XnPoint3D xnp){
	return Point((int)xnp.X,(int)xnp.Y);
}

void addTextToFrame(Mat &frame, int x, int y, const string text){
	putText(frame, text, Point(x,y), CV_FONT_NORMAL, 0.8, CV_RGB(255,255,255));
}

void DrawJoints(Mat &userMask)
{
	int nUsers = g_UserHandler.getUserCount();
	if(nUsers == 0){
		return;
	}

	for(int i=0;i<nUsers;i++){

		User u = g_UserHandler.getUser(i);
		XnUserID uid = u.getUserID();

		if(uid==0) continue; // the retrieved user is a false one

		// user is not calibrated
		if (!g_UserGenerator.GetSkeletonCap().IsCalibrated(uid)){
			//cout << "user not calibrated " << uid << endl;
			continue;
		}
		if (!g_UserGenerator.GetSkeletonCap().IsTracking(uid)){
			//cout << "user not tracked " << uid << endl;
			continue;
		}

		//cout << "fetching userpoints" << endl;
		UserPoints up = u.getLatestUserPoints();

		XnPoint3D pt[6];
		pt[0] = up.getLShoulder();
		pt[1] = up.getRShoulder();
		pt[2] = up.getLHip();
		pt[3] = up.getRHip();
		pt[4] = up.getTorso();
		pt[5] = up.calcMiddlePoint(pt[0], pt[1]);

		g_DepthGenerator.ConvertRealWorldToProjective(6, pt, pt);

		Point cvPt1 = Xn2CVPoint(pt[0]);
		Point cvPt2 = Xn2CVPoint(pt[1]);
		Point cvPt3 = Xn2CVPoint(pt[2]);
		Point cvPt4 = Xn2CVPoint(pt[3]);
		Point cvPt5 = Xn2CVPoint(pt[4]);
		Point cvPt6 = Xn2CVPoint(pt[5]);
		circle(userMask,cvPt1, 5, CV_RGB(255,255,255));
		circle(userMask,cvPt2, 5, CV_RGB(255,255,255));
		circle(userMask,cvPt3, 5, CV_RGB(255,255,255));
		circle(userMask,cvPt4, 5, CV_RGB(255,255,255));
		circle(userMask,cvPt5, 10, CV_RGB(200,200,200), -1);
		circle(userMask,cvPt6, 15, CV_RGB(150,150,150), -1);
		line(userMask, cvPt1, cvPt2, CV_RGB(255,255,255));
		line(userMask, cvPt3, cvPt4, CV_RGB(255,255,255));
		line(userMask, cvPt1, cvPt3, CV_RGB(255,255,255));
		line(userMask, cvPt2, cvPt4, CV_RGB(255,255,255));
		line(userMask, cvPt1, cvPt4, CV_RGB(255,255,255));
		line(userMask, cvPt2, cvPt3, CV_RGB(255,255,255));
	}
}

void highlightPlayers(Mat &userMask, XnUserID openingDoor){
	XnUserID aUsers[20];
	XnUInt16 nUsers = 20;

    g_UserGenerator.GetUsers(aUsers, nUsers);
    xn::SceneMetaData sceneMD;
    g_UserGenerator.GetUserPixels(0,sceneMD);
    const XnLabel* pLabels = sceneMD.Data();

    for (int y = 0; y < 480; ++y){
    	for (int x = 0; x < 640; ++x, ++pLabels){
    		int label = (*pLabels);
    		if(label != 0){
    			if(label == openingDoor)
    				userMask.at<uchar>(y,x) = 200;
    			else
    				userMask.at<uchar>(y,x) = 128;
    		}else{
    			userMask.at<uchar>(y,x) = 0;
    		}
    	}
    }

   	DrawJoints(userMask);

}

void writingOverlay(Mat &layer){
	stringstream ss;
	ss << "# users: " << g_UserHandler.getUserCount();
	addTextToFrame(layer, 10, 20, ss.str());
	ss.str("");
	int nUsers = g_UserHandler.getUserCount();
	if(nUsers > 0){
		User u = g_UserHandler.getUser(0);

		if(u.countFeatureSets() > 0 && u.countUserPoints() > 3){
			FeatureSet fs = u.getLatestFeatureSet();
			UserPoints up = u.getLatestUserPoints();
			XnPoint3D torso = up.getTorso();

			ss << setprecision(3) << "av: " << fs.angular_velocity << " deg/s";
			addTextToFrame(layer, 10, 360, ss.str());
			ss.str("");
			ss << "acc: " << fs.acceleration << " m/s^2";
			addTextToFrame(layer, 10, 340, ss.str());
			ss.str("");
			ss << setprecision(4) << "torso: (" << torso.X << ", " << torso.Y << ", " << torso.Z << ")";
			addTextToFrame(layer, 10, 380, ss.str());
			ss.str("");
			ss << setprecision(3) << "ba: " << fs.body_alignment << " deg";
			addTextToFrame(layer, 10, 420, ss.str());
			ss.str("");
			ss << "v: " << fs.speed << " m/s";
			addTextToFrame(layer, 10, 440, ss.str());
			ss.str("");
			ss << "head: " << fs.heading << " deg";
			addTextToFrame(layer, 10, 460, ss.str());
			ss.str("");
			ss << "dist: " << fs.distance/1000 << " m";
			addTextToFrame(layer, 10, 400, ss.str());
		}
	}

}

void drawAngle(Mat &layer){
	if(g_UserHandler.getUserCount() > 0){
		User u = g_UserHandler.getUser(0);
		FeatureSet fs = u.getLatestFeatureSet();
		Point o1 = Point(610,40);
		Point o2 = Point(560,40);
		int r = 15;
		double a = fs.body_alignment;
		double b = fs.heading;
		if(a <= 180 && a >=0){
			double ar = (a/180) * 3.141592654;
			Point p1 = Point(o1.x-(sin(ar)*r), o1.y-(cos(ar)*r));
			circle(layer,o1, r, CV_RGB(255,255,255));
			line(layer, o1, p1, CV_RGB(150,150,150));
		}

		if(b <= 360 && b >=0){
			double br = (b/180) * 3.141592654;
			Point p2 = Point(o2.x-(sin(br)*r), o2.y-(cos(br)*r));
			circle(layer,o2, r, CV_RGB(255,255,255));
			line(layer, o2, p2, CV_RGB(150,150,150));
		}

	}
}

void PrintRC(XnStatus rc, const char *what){
	printf("%s failed: %s\n", what, xnGetStatusString(rc));
}

int main(){

	xn::EnumerationErrors errors;
	XnStatus nRetVal = XN_STATUS_OK;


	nRetVal = g_Context.InitFromXmlFile(SAMPLE_XML_PATH, &errors);
	if (nRetVal == XN_STATUS_NO_NODE_PRESENT)
	{
		XnChar strError[1024];
		errors.ToString(strError, 1024);
		printf("%s\n", strError);
		return (nRetVal);
	}
	else if (nRetVal != XN_STATUS_OK)
	{
		printf("Open failed: %s\n", xnGetStatusString(nRetVal));
		return (nRetVal);
	}

	nRetVal = g_Context.FindExistingNode(XN_NODE_TYPE_DEPTH, g_DepthGenerator);
	CHECK_RC(nRetVal, "Find depth generator");
	nRetVal = g_Context.FindExistingNode(XN_NODE_TYPE_USER, g_UserGenerator);
	if (nRetVal != XN_STATUS_OK)
	{
		nRetVal = g_UserGenerator.Create(g_Context);
		CHECK_RC(nRetVal, "Find user generator");
	}

	XnCallbackHandle hUserCallbacks, hCalibrationCallbacks, hPoseCallbacks;
	if (!g_UserGenerator.IsCapabilitySupported(XN_CAPABILITY_SKELETON))
	{
		printf("Supplied user generator doesn't support skeleton\n");
		return 1;
	}
	g_UserGenerator.RegisterUserCallbacks(User_NewUser, User_Exit, NULL, hUserCallbacks);
	g_UserGenerator.RegisterToUserExit(User_Exit, NULL, hUserCallbacks);
	g_UserGenerator.RegisterToUserReEnter(User_ReEnter, NULL, hUserCallbacks);
	g_UserGenerator.GetSkeletonCap().RegisterCalibrationCallbacks(UserCalibration_CalibrationStart, UserCalibration_CalibrationEnd, NULL, hCalibrationCallbacks);

	if (g_UserGenerator.GetSkeletonCap().NeedPoseForCalibration())
	{
		g_bNeedPose = TRUE;
		if (!g_UserGenerator.IsCapabilitySupported(XN_CAPABILITY_POSE_DETECTION))
		{
			printf("Pose required, but not supported\n");
			return 1;
		}
		g_UserGenerator.GetPoseDetectionCap().RegisterToPoseCallbacks(UserPose_PoseDetected, NULL, NULL, hPoseCallbacks);
		g_UserGenerator.GetSkeletonCap().GetCalibrationPose(g_strPose);
	}

	g_UserGenerator.GetSkeletonCap().SetSkeletonProfile(XN_SKEL_PROFILE_ALL);

	nRetVal = g_Context.StartGeneratingAll();
	CHECK_RC(nRetVal, "StartGenerating");

	xn::DepthMetaData depthMD;
	xn::ImageMetaData imageMD;

	cv::Mat cvDepth;
	cv::Mat cvImage;
	cv::Mat cvUsers(480, 640, CV_8UC1, cv::Scalar::all(0));
	// main loop

	namedWindow("Users", CV_WINDOW_AUTOSIZE);
	cvMoveWindow("Users", 300, 300);
	ofstream myfile;
	myfile.open("data.txt");

	int grid = -1;

	while(!g_bQuit){

		// wait for new data to be available
		nRetVal = g_Context.WaitAndUpdateAll();
		CHECK_RC(nRetVal, "Updating sensor data");

		XnUserID openingDoor;

		g_UserHandler.updateUserPoints();
		g_UserHandler.runUserCalculations();
		g_UserHandler.inferUserIntentions(openingDoor);

		highlightPlayers(cvUsers, openingDoor);
		writingOverlay(cvUsers);
		drawAngle(cvUsers);

		User u = g_UserHandler.getUser(0);

		if(u.countFeatureSets() > 0 && u.countUserPoints() > 3){
			FeatureSet fs = u.getLatestFeatureSet();
			UserPoints up = u.getLatestUserPoints();
			XnPoint3D torso = up.getTorso();
			
			setprecision(4);
			int xpos = torso.X;
			int ypos = torso.Y;

			xpos += 800;

			if (xpos < 0 || xpos >= 1600 || ypos < 0 || ypos >= 800) {
				myfile << -1 << ":" << u.getUserID() << ";" ;
			} else {
				int xGrid = xpos / 200;
				int yGrid = ypos / 100;
		
				int coor = ((7 - yGrid) * 8) + xGrid;

				if (coor != grid) {
					myfile << coor << ":" << u.getUserID() << ";" ;
					grid = coor;
				}
			}
		}
		imshow("Users", cvUsers);

		unsigned char k = waitKey(30);

		if(k > 0){
			switch(k){
			case 's':
				cout << "saving calibration ..." << endl;
				SaveCalibration();
				break;
			case 'q':
			case 'Q':
			case 27:
				g_bQuit = TRUE;
				break;
			}
		}
	}

	myfile.close();

	CleanupExit();



}
